# This file is generated by codegen.py. DO NOT EDIT!

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Callable, List, Optional

# 17: override Comment
# 32: override Diagram
# 29: override PackageableElement
# 26: override NamedElement
# 14: override Element
from gaphor.core.modeling import (
    Comment,
    Diagram,
    Element,
    NamedElement,
    PackageableElement,
    StyleSheet,
)
from gaphor.core.modeling.properties import (
    association,
    attribute,
    derived,
    derivedunion,
    enumeration,
    redefine,
    relation_many,
    relation_one,
)


class InstanceSpecification(PackageableElement):
    specification: attribute[str]
    slot: relation_many[Slot]
    classifier: relation_many[Classifier]
    extended: relation_many[Element]


class EnumerationLiteral(InstanceSpecification):
    enumeration: relation_one[Enumeration]


class Relationship(Element):
    relatedElement: relation_many[Element]


class DirectedRelationship(Relationship):
    target: relation_many[Element]
    source: relation_many[Element]


class PackageMerge(DirectedRelationship):
    mergingPackage: relation_one[Package]
    mergedPackage: relation_one[Package]


class Namespace(NamedElement):
    ownedRule: relation_many[Constraint]
    elementImport: relation_many[ElementImport]
    packageImport: relation_many[PackageImport]
    ownedMember: relation_many[NamedElement]
    member: relation_many[NamedElement]
    importedMember: derivedunion[PackageableElement]


class Type(PackageableElement):
    package: relation_one[Package]


class RedefinableElement(NamedElement):
    isLeaf: attribute[int]
    visibility: enumeration
    redefinedElement: relation_many[RedefinableElement]
    redefinitionContext: relation_many[Classifier]


class Classifier(Namespace, Type, RedefinableElement):
    isAbstract: attribute[int]
    ownedUseCase: relation_many[UseCase]
    generalization: relation_many[Generalization]
    useCase: relation_many[UseCase]
    redefinedClassifier: relation_many[Classifier]
    substitution: relation_many[Substitution]
    attribute: relation_many[Property]
    feature: relation_many[Feature]
    general: derived[Classifier]
    inheritedMember: derivedunion[NamedElement]


class Association(Classifier, Relationship):
    isDerived: attribute[int]
    memberEnd: relation_many[Property]
    ownedEnd: relation_many[Property]
    navigableOwnedEnd: relation_many[Property]
    endType: derived[Type]


class Extension(Association):
    isRequired: attribute[int]
    ownedEnd: relation_one[ExtensionEnd]  # type: ignore[assignment]
    metaclass: property


class Actor(Classifier):
    ownedAttribute: relation_many[Property]


class ActivityNode(RedefinableElement):
    outgoing: relation_many[ActivityEdge]
    incoming: relation_many[ActivityEdge]
    inGroup: relation_many[ActivityGroup]
    inPartition: relation_many[ActivityPartition]
    redefinedElement: relation_many[ActivityNode]  # type: ignore[assignment]


class ControlNode(ActivityNode):
    pass


class MergeNode(ControlNode):
    pass


class Feature(RedefinableElement):
    isStatic: attribute[int]
    featuringClassifier: relation_many[Classifier]


class ActivityEdge(RedefinableElement):
    activity: relation_one[Activity]
    guard: attribute[str]
    source: relation_one[ActivityNode]
    target: relation_one[ActivityNode]
    inGroup: relation_many[ActivityGroup]
    redefinedElement: relation_many[ActivityEdge]  # type: ignore[assignment]


class ObjectFlow(ActivityEdge):
    pass


class FinalNode(ControlNode):
    pass


class ActivityFinalNode(FinalNode):
    pass


class CommunicationPath(Association):
    pass


class Dependency(DirectedRelationship, PackageableElement):
    client: relation_many[NamedElement]
    supplier: relation_many[NamedElement]


class Permission(Dependency):
    pass


class Abstraction(Dependency):
    mapping: attribute[str]


class Realization(Abstraction):
    realizingClassifier: relation_one[Classifier]
    abstraction: relation_one[Component]


class TypedElement(NamedElement):
    type: relation_one[Type]
    typeValue: attribute[str]


class ObjectNode(TypedElement, ActivityNode):
    ordering: enumeration
    isControlType: attribute[int]
    upperBound: attribute[str]
    selection: relation_one[Behavior]


class Pin(ObjectNode):
    pass


class Generalization(DirectedRelationship):
    isSubstitutable: attribute[int]
    general: relation_one[Classifier]
    specific: relation_one[Classifier]


class BehavioredClassifier(Classifier):
    ownedBehavior: relation_many[Behavior]
    implementation: relation_many[Implementation]  # type: ignore[assignment]


class StructuredClassifier(Classifier):
    ownedConnector: relation_many[Connector]
    ownedAttribute: relation_many[Property]
    role: relation_many[ConnectableElement]
    part: property


class EncapsulatedClassifer(StructuredClassifier):
    ownedPort: relation_many[Port]


class Class(BehavioredClassifier, EncapsulatedClassifer):
    ownedOperation: relation_many[Operation]
    nestedClassifier: relation_many[Classifier]
    ownedAttribute: relation_many[Property]
    ownedReception: relation_many[Reception]
    extension: property
    superClass: derived[Classifier]


class DeploymentTarget(NamedElement):
    deployment: relation_many[Deployment]


class Node(Class, DeploymentTarget):
    nestedNode: relation_many[Node]


class Device(Node):
    pass


class MultiplicityElement(Element):
    isUnique: attribute[int]
    isOrdered: attribute[int]
    upperValue: attribute[str]
    lowerValue: attribute[str]
    lower: attribute[str]
    upper: attribute[str]


class StructuralFeature(MultiplicityElement, TypedElement, Feature):
    isReadOnly: attribute[int]
    slot: relation_many[Slot]


class UseCase(BehavioredClassifier):
    subject: relation_many[Classifier]
    extensionPoint: relation_many[ExtensionPoint]
    include: relation_many[Include]
    extend: relation_many[Extend]
    ownedAttribute: relation_many[Property]


class InputPin(Pin):
    pass


class Manifestation(Abstraction):
    pass


class Component(Class):
    isIndirectlyInstantiated: attribute[int]
    realization: relation_many[Realization]
    required: property
    provided: property
    ownedMember: relation_many[PackageableElement]  # type: ignore[assignment]


class ConnectableElement(TypedElement):
    end: relation_many[ConnectorEnd]


class Interface(Classifier, ConnectableElement):
    ownedAttribute: relation_many[Property]
    redefinedInterface: relation_many[Interface]
    nestedInterface: relation_many[Interface]
    ownedOperation: relation_many[Operation]
    ownedReception: relation_many[Reception]


class Include(DirectedRelationship):
    addition: relation_one[UseCase]
    includingCase: relation_one[UseCase]


class PackageImport(DirectedRelationship):
    visibility: enumeration
    importedPackage: relation_one[Package]
    importingNamespace: relation_one[Namespace]


class ProfileApplication(PackageImport):
    importedProfile: relation_one[Profile]


class ExtensionPoint(RedefinableElement):
    useCase: relation_one[UseCase]


class Usage(Dependency):
    pass


class ElementImport(DirectedRelationship):
    visibility: enumeration
    alias: attribute[str]
    importingNamespace: relation_one[Namespace]
    importedElement: relation_one[PackageableElement]


class Property(StructuralFeature, ConnectableElement):
    aggregation: enumeration
    isDerivedUnion: attribute[int]
    isDerived: attribute[int]
    isReadOnly: attribute[int]
    datatype: relation_one[DataType]
    subsettedProperty: relation_many[Property]
    classifier: relation_one[Classifier]
    redefinedProperty: relation_many[Property]
    class_: relation_one[Class]
    defaultValue: attribute[str]
    association: relation_one[Association]
    interface_: relation_one[Interface]
    owningAssociation: relation_one[Association]
    useCase: relation_one[UseCase]
    actor: relation_one[Actor]
    isComposite: derived[bool]
    navigability: derived[Optional[bool]]
    opposite: relation_one[Optional[Property]]


class ExtensionEnd(Property):
    type: relation_one[Stereotype]  # type: ignore[assignment]


class DataType(Classifier):
    ownedAttribute: relation_many[Property]
    ownedOperation: relation_many[Operation]


class Enumeration(DataType):
    literal: relation_many[EnumerationLiteral]


class Slot(Element):
    value: attribute[str]
    owningInstance: relation_one[InstanceSpecification]
    definingFeature: relation_one[StructuralFeature]


class ExecutableNode(ActivityNode):
    pass


class InitialNode(ControlNode):
    pass


class Stereotype(Class):
    pass


class DeployedArtifact(NamedElement):
    pass


class Artifact(Classifier, DeployedArtifact):
    manifestation: relation_many[Manifestation]


class ActivityParameterNode(ObjectNode):
    parameter: relation_one[Parameter]


class PrimitiveType(DataType):
    pass


class DecisionNode(ControlNode):
    decisionInput: relation_one[Behavior]


class Package(Namespace, PackageableElement):
    ownedDiagram: relation_many[Diagram]
    nestedPackage: relation_many[Package]
    package: relation_one[Package]
    ownedClassifier: relation_many[Type]
    packageExtension: relation_many[PackageMerge]
    appliedProfile: relation_many[ProfileApplication]
    ownedMember: relation_many[PackageableElement]  # type: ignore[assignment]


class Profile(Package):
    metamodelReference: relation_many[PackageImport]
    ownedStereotype: relation_many[Stereotype]
    metaclassReference: relation_many[ElementImport]


class Behavior(Class):
    isReentrant: attribute[int]
    redefinedBehavior: relation_many[Behavior]
    context: relation_one[BehavioredClassifier]


class Activity(Behavior):
    body: attribute[str]
    language: attribute[str]
    edge: relation_many[ActivityEdge]
    group: relation_many[ActivityGroup]
    node: relation_many[ActivityNode]
    action: relation_many[Action]


class Implementation(Realization):
    contract: relation_many[Interface]  # type: ignore[assignment]
    implementatingClassifier: relation_many[BehavioredClassifier]  # type: ignore[assignment]


class Parameter(TypedElement, MultiplicityElement):
    direction: enumeration
    defaultValue: attribute[str]
    ownerFormalParam: relation_one[BehavioralFeature]
    ownerReturnParam: relation_one[BehavioralFeature]
    operation: relation_one[Operation]  # type: ignore[assignment]


class BehavioralFeature(Feature, Namespace):
    isAbstract: attribute[int]
    method: relation_many[Behavior]
    formalParameter: relation_many[Parameter]
    raisedException: relation_many[Type]
    returnResult: relation_many[Parameter]
    parameter: relation_many[Parameter]


class Operation(BehavioralFeature):
    isQuery: attribute[int]
    precondition: relation_many[Constraint]
    bodyCondition: relation_one[Constraint]
    redefinedOperation: relation_many[Operation]
    class_: relation_one[Class]
    datatype: relation_one[DataType]
    postcondition: relation_many[Constraint]
    interface_: relation_one[Interface]
    raisedException: relation_many[Type]
    type: derivedunion[DataType]
    formalParameter: relation_many[Parameter]  # type: ignore[assignment]


class ControlFlow(ActivityEdge):
    pass


class Substitution(Realization):
    contract: relation_one[Classifier]
    substitutingClassifier: relation_one[Classifier]


class OutputPin(Pin):
    pass


class ValuePin(InputPin):
    value_: attribute[str]


class Action(ExecutableNode):
    effect: attribute[str]
    interaction: relation_one[Interaction]
    output: relation_many[OutputPin]
    context_: relation_one[Classifier]
    input: relation_many[InputPin]


class ExecutionEnvironment(Node):
    pass


class Extend(DirectedRelationship):
    extendedCase: relation_one[UseCase]
    extensionLocation: relation_many[ExtensionPoint]
    extension: relation_one[UseCase]
    constraint: relation_one[Constraint]


class ActivityGroup(Element):
    activity: relation_one[Activity]
    edgeContents: relation_many[ActivityEdge]
    nodeContents: relation_many[ActivityNode]
    superGroup: relation_one[ActivityGroup]
    subgroup: relation_many[ActivityGroup]


class Constraint(PackageableElement):
    constrainedElement: relation_many[Element]
    specification: attribute[str]
    stateInvariant: relation_one[StateInvariant]
    owningState: relation_one[State]
    parameterSet: relation_one[ParameterSet]
    context: derivedunion[Namespace]


class InteractionFragment(NamedElement):
    enclosingInteraction: relation_one[Interaction]
    covered: relation_one[Lifeline]
    generalOrdering: relation_many[GeneralOrdering]


class Interaction(Behavior, InteractionFragment):
    fragment: relation_many[InteractionFragment]
    lifeline: relation_many[Lifeline]
    message: relation_many[Message]
    action: relation_many[Action]


class StateInvariant(InteractionFragment):
    invariant: relation_one[Constraint]
    covered: relation_one[Lifeline]  # type: ignore[assignment]


class Lifeline(NamedElement):
    coveredBy: relation_many[InteractionFragment]
    interaction: relation_one[Interaction]
    parse: Callable[[Lifeline, str], None]
    render: Callable[[Lifeline], str]


class Message(NamedElement):
    messageKind: property
    messageSort: enumeration
    argument: attribute[str]
    sendEvent: relation_one[MessageEnd]
    receiveEvent: relation_one[MessageEnd]
    interaction: relation_one[Interaction]
    signature: relation_one[NamedElement]


class MessageEnd(NamedElement):
    sendMessage: relation_one[Message]
    receiveMessage: relation_one[Message]


class OccurrenceSpecification(InteractionFragment):
    covered: relation_one[Lifeline]  # type: ignore[assignment]


class GeneralOrdering(NamedElement):
    interactionFragment: relation_one[InteractionFragment]


class Connector(Feature):
    kind: enumeration
    redefinedConnector: relation_many[Connector]
    type: relation_one[Association]
    end: relation_many[ConnectorEnd]
    contract: relation_many[Behavior]


class ConnectorEnd(MultiplicityElement):
    role: relation_one[ConnectableElement]
    partWithPort: relation_one[Property]
    definingEnd: relation_one[Property]


class FlowFinalNode(FinalNode):
    pass


class JoinNode(ControlNode):
    isCombineDuplicate: attribute[int]
    joinSpec: attribute[str]


class ForkNode(ControlNode):
    pass


class StateMachine(Behavior):
    region: relation_many[Region]
    extendedStateMachine: relation_one[StateMachine]


class Region(Namespace, RedefinableElement):
    stateMachine: relation_one[StateMachine]
    subvertex: relation_many[Vertex]
    state: relation_one[State]
    extendedRegion: relation_many[Region]  # type: ignore[assignment]


# 35: override Transition
# Invert order of superclasses to avoid MRO issues
class Transition(RedefinableElement, NamedElement):
    kind: enumeration
    container: relation_one[Region]
    source: relation_one[Vertex]
    target: relation_one[Vertex]
    effect: relation_one[Behavior]
    guard: relation_one[Constraint]
    redefinitionContext: relation_many[Classifier]
    redefinedTransition: relation_many[Transition]


class Vertex(NamedElement):
    container: relation_one[Region]
    outgoing: relation_many[Transition]
    incoming: relation_many[Transition]


class Pseudostate(Vertex):
    kind: enumeration
    stateMachine: relation_one[StateMachine]
    state: relation_one[State]


class ConnectionPointReference(Vertex):
    entry: relation_many[Pseudostate]
    exit: relation_many[Pseudostate]
    state: relation_one[State]


class State(Vertex, Namespace, RedefinableElement):
    entry: relation_one[Behavior]
    exit: relation_one[Behavior]
    doActivity: relation_one[Behavior]
    statevariant: relation_one[Constraint]
    submachine: relation_one[StateMachine]
    redefinedState: relation_many[State]  # type: ignore[assignment]


class FinalState(State):
    pass


class Port(Property):
    isBehavior: attribute[int]
    isService: attribute[int]
    encapsulatedClassifier: relation_one[EncapsulatedClassifer]


class Deployment(Dependency):
    deployedArtifact: relation_many[DeployedArtifact]


class ActivityPartition(ActivityGroup, NamedElement):
    isDimension: attribute[int]
    isExternal: attribute[int]
    node: relation_many[ActivityNode]
    represents: relation_one[Element]
    subpartition: relation_many[ActivityPartition]


class MessageOccurrenceSpecification(MessageEnd, OccurrenceSpecification):
    pass


class AcceptEventAction(Action):
    isUnmarshall: attribute[int]
    result: relation_many[OutputPin]


class ReplyAction(Action):
    replyValue: relation_one[InputPin]
    returnInformation: relation_one[InputPin]


class UnmarshallAction(Action):
    result: relation_many[OutputPin]
    unmarshallType: relation_one[Classifier]
    object: relation_one[InputPin]


class AcceptCallAction(AcceptEventAction):
    returnInformation: relation_one[OutputPin]


class InvocationAction(Action):
    pass


class SendSignalAction(InvocationAction):
    target: relation_many[InputPin]


class Collaboration(StructuredClassifier, BehavioredClassifier):
    collaborationRole: relation_many[ConnectableElement]


class Trigger(NamedElement):
    event: relation_one[Event]


class Event(PackageableElement):
    pass


class Signal(Classifier):
    ownedAttribute: relation_many[Property]


class Reception(BehavioralFeature):
    signal: relation_one[Signal]


class ExecutionSpecification(InteractionFragment):
    executionOccurrenceSpecification: relation_many[ExecutionOccurrenceSpecification]
    start: relation_one[ExecutionOccurrenceSpecification]
    finish: relation_one[ExecutionOccurrenceSpecification]


class ExecutionOccurrenceSpecification(OccurrenceSpecification):
    execution: relation_one[ExecutionSpecification]


class ActionExecutionSpecification(ExecutionSpecification):
    action: relation_one[Action]


class BehaviorExecutionSpecification(ExecutionSpecification):
    behavior: relation_one[Behavior]


class ChangeEvent(Event):
    changeExpression: attribute[str]


class StructuralFeatureAction(Action):
    pass


class WriteStructuralFeatureAction(StructuralFeatureAction):
    pass


class AddStructuralFeatureValueAction(WriteStructuralFeatureAction):
    isReplaceAll: attribute[int]


class ParameterSet(NamedElement):
    condition: relation_many[Constraint]


# class 'ValueSpecification' has been stereotyped as 'SimpleAttribute'
# class 'InstanceValue' has been stereotyped as 'SimpleAttribute' too
# class 'Expression' has been stereotyped as 'SimpleAttribute' too
# class 'LiteralSpecification' has been stereotyped as 'SimpleAttribute' too
# class 'LiteralUnlimitedNatural' has been stereotyped as 'SimpleAttribute' too
# class 'LiteralBoolean' has been stereotyped as 'SimpleAttribute' too
# class 'LiteralInteger' has been stereotyped as 'SimpleAttribute' too
# class 'LiteralString' has been stereotyped as 'SimpleAttribute' too
# class 'LiteralNull' has been stereotyped as 'SimpleAttribute' too
# class 'OpaqueExpression' has been stereotyped as 'SimpleAttribute' too
Extension.isRequired = attribute("isRequired", int)
Feature.isStatic = attribute("isStatic", int, default=False)
RedefinableElement.isLeaf = attribute("isLeaf", int, default=True)
RedefinableElement.visibility = enumeration(
    "visibility", ("public", "private", "package", "protected"), "public"
)
Generalization.isSubstitutable = attribute("isSubstitutable", int)
ObjectNode.ordering = enumeration(
    "ordering", ("unordered", "ordered", "LIFO", "FIFO"), "FIFO"
)
ObjectNode.isControlType = attribute("isControlType", int, default=False)
StructuralFeature.isReadOnly = attribute("isReadOnly", int, default=False)
Component.isIndirectlyInstantiated = attribute(
    "isIndirectlyInstantiated", int, default=True
)
Association.isDerived = attribute("isDerived", int, default=False)
ElementImport.visibility = enumeration(
    "visibility", ("public", "private", "package", "protected"), "public"
)
ElementImport.alias = attribute("alias", str)
MultiplicityElement.isUnique = attribute("isUnique", int, default=True)
MultiplicityElement.isOrdered = attribute("isOrdered", int, default=True)
Activity.body = attribute("body", str)
Activity.language = attribute("language", str)
Classifier.isAbstract = attribute("isAbstract", int, default=False)
Parameter.direction = enumeration("direction", ("inout", "in", "out", "return"), "in")
Operation.isQuery = attribute("isQuery", int, default=False)
Property.aggregation = enumeration(
    "aggregation", ("none", "shared", "composite"), "none"
)
Property.isDerivedUnion = attribute("isDerivedUnion", int, default=False)
Property.isDerived = attribute("isDerived", int, default=False)
Property.isReadOnly = attribute("isReadOnly", int, default=False)
Behavior.isReentrant = attribute("isReentrant", int)
BehavioralFeature.isAbstract = attribute("isAbstract", int)
Action.effect = attribute("effect", str)
PackageImport.visibility = enumeration(
    "visibility", ("public", "private", "package", "protected"), "public"
)
# 115: override Message.messageKind: property
# defined in umloverrides.py

Message.messageSort = enumeration(
    "messageSort",
    (
        "synchCall",
        "asynchCall",
        "asynchSignal",
        "createMessage",
        "deleteMessage",
        "reply",
    ),
    "synchCall",
)
Connector.kind = enumeration("kind", ("assembly", "delegation"), "assembly")
JoinNode.isCombineDuplicate = attribute("isCombineDuplicate", int, default=True)
Transition.kind = enumeration("kind", ("internal", "local", "external"), "internal")
Pseudostate.kind = enumeration(
    "kind",
    (
        "initial",
        "deepHistory",
        "shallowHistory",
        "join",
        "fork",
        "junction",
        "choice",
        "entryPoint",
        "exitPoint",
        "terminate",
    ),
    "initial",
)
Port.isBehavior = attribute("isBehavior", int)
Port.isService = attribute("isService", int)
ActivityPartition.isDimension = attribute("isDimension", int, default=False)
ActivityPartition.isExternal = attribute("isExternal", int, default=False)
AcceptEventAction.isUnmarshall = attribute("isUnmarshall", int, default=False)
AddStructuralFeatureValueAction.isReplaceAll = attribute(
    "isReplaceAll", int, default=False
)
Operation.precondition = association("precondition", Constraint, composite=True)
Package.ownedDiagram = association(
    "ownedDiagram", Diagram, composite=True, opposite="package"
)
Diagram.package = association("package", Package, upper=1, opposite="ownedDiagram")
Package.nestedPackage = association(
    "nestedPackage", Package, composite=True, opposite="package"
)
Package.package = association("package", Package, upper=1, opposite="nestedPackage")
NamedElement.clientDependency = association(
    "clientDependency", Dependency, opposite="client"
)
Dependency.client = association(
    "client", NamedElement, lower=1, opposite="clientDependency"
)
DecisionNode.decisionInput = association("decisionInput", Behavior, upper=1)
Activity.edge = association("edge", ActivityEdge, composite=True, opposite="activity")
ActivityEdge.activity = association("activity", Activity, upper=1, opposite="edge")
Substitution.contract = association("contract", Classifier, lower=1, upper=1)
Operation.bodyCondition = association(
    "bodyCondition", Constraint, upper=1, composite=True
)
# 'InstanceSpecification.specification' is a simple attribute
InstanceSpecification.specification = attribute("specification", str)
BehavioralFeature.method = association("method", Behavior)
Property.datatype = association(
    "datatype", DataType, upper=1, opposite="ownedAttribute"
)
DataType.ownedAttribute = association(
    "ownedAttribute", Property, composite=True, opposite="datatype"
)
TypedElement.type = association("type", Type, upper=1)
ActivityParameterNode.parameter = association("parameter", Parameter, lower=1, upper=1)
Dependency.supplier = association(
    "supplier", NamedElement, lower=1, opposite="supplierDependency"
)
NamedElement.supplierDependency = association(
    "supplierDependency", Dependency, opposite="supplier"
)
Operation.redefinedOperation = association("redefinedOperation", Operation)
Activity.group = association(
    "group", ActivityGroup, composite=True, opposite="activity"
)
ActivityGroup.activity = association("activity", Activity, upper=1, opposite="group")
Package.ownedClassifier = association(
    "ownedClassifier", Type, composite=True, opposite="package"
)
Type.package = association("package", Package, upper=1, opposite="ownedClassifier")
Property.subsettedProperty = association("subsettedProperty", Property)
Property.classifier = association(
    "classifier", Classifier, upper=1, opposite="attribute"
)
Profile.metamodelReference = association(
    "metamodelReference", PackageImport, composite=True
)
# 'ActivityEdge.guard' is a simple attribute
ActivityEdge.guard = attribute("guard", str)
Class.ownedOperation = association(
    "ownedOperation", Operation, composite=True, opposite="class_"
)
Operation.class_ = association("class_", Class, upper=1, opposite="ownedOperation")
Enumeration.literal = association(
    "literal", EnumerationLiteral, composite=True, opposite="enumeration"
)
EnumerationLiteral.enumeration = association(
    "enumeration", Enumeration, upper=1, opposite="literal"
)
ActivityEdge.source = association(
    "source", ActivityNode, lower=1, upper=1, opposite="outgoing"
)
ActivityNode.outgoing = association("outgoing", ActivityEdge, opposite="source")
Profile.ownedStereotype = association("ownedStereotype", Stereotype, composite=True)
Property.redefinedProperty = association("redefinedProperty", Property)
DataType.ownedOperation = association(
    "ownedOperation", Operation, composite=True, opposite="datatype"
)
Operation.datatype = association(
    "datatype", DataType, upper=1, opposite="ownedOperation"
)
Generalization.general = association("general", Classifier, lower=1, upper=1)
Classifier.ownedUseCase = association("ownedUseCase", UseCase, composite=True)
# 'MultiplicityElement.upperValue' is a simple attribute
MultiplicityElement.upperValue = attribute("upperValue", str)
PackageMerge.mergingPackage = association(
    "mergingPackage", Package, lower=1, upper=1, opposite="packageExtension"
)
Package.packageExtension = association(
    "packageExtension", PackageMerge, composite=True, opposite="mergingPackage"
)
Package.appliedProfile = association(
    "appliedProfile", ProfileApplication, composite=True
)
Activity.node = association("node", ActivityNode, composite=True)
# 'Parameter.defaultValue' is a simple attribute
Parameter.defaultValue = attribute("defaultValue", str)
Class.nestedClassifier = association("nestedClassifier", Classifier, composite=True)
# 'Slot.value' is a simple attribute
Slot.value = attribute("value", str)
Include.addition = association("addition", UseCase, lower=1, upper=1)
Realization.realizingClassifier = association(
    "realizingClassifier", Classifier, lower=1, upper=1
)
# 'TypedElement.typeValue' is a simple attribute
TypedElement.typeValue = attribute("typeValue", str)
Constraint.constrainedElement = association("constrainedElement", Element)
PackageMerge.mergedPackage = association("mergedPackage", Package, lower=1, upper=1)
BehavioralFeature.formalParameter = association(
    "formalParameter", Parameter, composite=True, opposite="ownerFormalParam"
)
Parameter.ownerFormalParam = association(
    "ownerFormalParam", BehavioralFeature, upper=1, opposite="formalParameter"
)
Class.ownedAttribute = association(
    "ownedAttribute", Property, composite=True, opposite="class_"
)
Property.class_ = association("class_", Class, upper=1, opposite="ownedAttribute")
Extend.extendedCase = association("extendedCase", UseCase, lower=1, upper=1)
# 'Property.defaultValue' is a simple attribute
Property.defaultValue = attribute("defaultValue", str)
Namespace.ownedRule = association("ownedRule", Constraint, composite=True)
Property.association = association(
    "association", Association, upper=1, opposite="memberEnd"
)
Association.memberEnd = association(
    "memberEnd", Property, lower=2, composite=True, opposite="association"
)
Classifier.generalization = association(
    "generalization", Generalization, composite=True, opposite="specific"
)
Generalization.specific = association(
    "specific", Classifier, lower=1, upper=1, opposite="generalization"
)
Realization.abstraction = association(
    "abstraction", Component, upper=1, opposite="realization"
)
Component.realization = association(
    "realization", Realization, composite=True, opposite="abstraction"
)
# 'ValuePin.value_' is a simple attribute
ValuePin.value_ = attribute("value_", str)
BehavioralFeature.raisedException = association("raisedException", Type)
Activity.action = association("action", Action, composite=True)
# 'Abstraction.mapping' is a simple attribute
Abstraction.mapping = attribute("mapping", str)
ActivityNode.incoming = association("incoming", ActivityEdge, opposite="target")
ActivityEdge.target = association(
    "target", ActivityNode, lower=1, upper=1, opposite="incoming"
)
Extend.extensionLocation = association("extensionLocation", ExtensionPoint, lower=1)
Property.interface_ = association(
    "interface_", Interface, upper=1, opposite="ownedAttribute"
)
Interface.ownedAttribute = association(
    "ownedAttribute", Property, composite=True, opposite="interface_"
)
ActivityGroup.edgeContents = association(
    "edgeContents", ActivityEdge, opposite="inGroup"
)
ActivityEdge.inGroup = association("inGroup", ActivityGroup, opposite="edgeContents")
Slot.owningInstance = association(
    "owningInstance", InstanceSpecification, lower=1, upper=1, opposite="slot"
)
InstanceSpecification.slot = association(
    "slot", Slot, composite=True, opposite="owningInstance"
)
UseCase.subject = association("subject", Classifier, opposite="useCase")
Classifier.useCase = association("useCase", UseCase, opposite="subject")
Property.owningAssociation = association(
    "owningAssociation", Association, upper=1, opposite="ownedEnd"
)
Association.ownedEnd = association(
    "ownedEnd", Property, composite=True, opposite="owningAssociation"
)
Interface.redefinedInterface = association("redefinedInterface", Interface)
Artifact.manifestation = association("manifestation", Manifestation, composite=True)
ExtensionPoint.useCase = association(
    "useCase", UseCase, lower=1, upper=1, opposite="extensionPoint"
)
UseCase.extensionPoint = association(
    "extensionPoint", ExtensionPoint, opposite="useCase"
)
Operation.postcondition = association("postcondition", Constraint, composite=True)
Extension.ownedEnd = association(
    "ownedEnd", ExtensionEnd, lower=1, upper=1, composite=True
)
# 'Constraint.specification' is a simple attribute
Constraint.specification = attribute("specification", str)
Profile.metaclassReference = association(
    "metaclassReference", ElementImport, composite=True
)
Namespace.elementImport = association(
    "elementImport", ElementImport, composite=True, opposite="importingNamespace"
)
ElementImport.importingNamespace = association(
    "importingNamespace", Namespace, upper=1, opposite="elementImport"
)
# 'MultiplicityElement.lowerValue' is a simple attribute
MultiplicityElement.lowerValue = attribute("lowerValue", str)
Interface.nestedInterface = association("nestedInterface", Interface, composite=True)
InstanceSpecification.classifier = association("classifier", Classifier)
Interface.ownedOperation = association(
    "ownedOperation", Operation, composite=True, opposite="interface_"
)
Operation.interface_ = association(
    "interface_", Interface, upper=1, opposite="ownedOperation"
)
ElementImport.importedElement = association(
    "importedElement", PackageableElement, lower=1, upper=1
)
Parameter.ownerReturnParam = association(
    "ownerReturnParam", BehavioralFeature, upper=1, opposite="returnResult"
)
BehavioralFeature.returnResult = association(
    "returnResult", Parameter, composite=True, opposite="ownerReturnParam"
)
Classifier.redefinedClassifier = association("redefinedClassifier", Classifier)
Substitution.substitutingClassifier = association(
    "substitutingClassifier", Classifier, lower=1, upper=1, opposite="substitution"
)
Classifier.substitution = association(
    "substitution", Substitution, composite=True, opposite="substitutingClassifier"
)
Operation.raisedException = association("raisedException", Type)
PackageImport.importedPackage = association(
    "importedPackage", Package, lower=1, upper=1
)
StructuralFeature.slot = association(
    "slot", Slot, composite=True, opposite="definingFeature"
)
Slot.definingFeature = association(
    "definingFeature", StructuralFeature, lower=1, upper=1, opposite="slot"
)
Include.includingCase = association(
    "includingCase", UseCase, lower=1, upper=1, opposite="include"
)
UseCase.include = association(
    "include", Include, composite=True, opposite="includingCase"
)
Extend.extension = association(
    "extension", UseCase, lower=1, upper=1, opposite="extend"
)
UseCase.extend = association("extend", Extend, composite=True, opposite="extension")
Extend.constraint = association("constraint", Constraint, upper=1, composite=True)
ProfileApplication.importedProfile = association(
    "importedProfile", Profile, lower=1, upper=1
)
Namespace.packageImport = association(
    "packageImport", PackageImport, composite=True, opposite="importingNamespace"
)
PackageImport.importingNamespace = association(
    "importingNamespace", Namespace, upper=1, opposite="packageImport"
)
Behavior.redefinedBehavior = association("redefinedBehavior", Behavior)
# 20: override Element.ownedComment
# Defined in Core model

# 23: override Comment.annotatedElement
# Defined in Core model

Behavior.context = association(
    "context", BehavioredClassifier, upper=1, opposite="ownedBehavior"
)
BehavioredClassifier.ownedBehavior = association(
    "ownedBehavior", Behavior, composite=True, opposite="context"
)
ActivityGroup.nodeContents = association(
    "nodeContents", ActivityNode, opposite="inGroup"
)
ActivityNode.inGroup = association("inGroup", ActivityGroup, opposite="nodeContents")
UseCase.ownedAttribute = association(
    "ownedAttribute", Property, composite=True, opposite="useCase"
)
Property.useCase = association("useCase", UseCase, upper=1, opposite="ownedAttribute")
Property.actor = association("actor", Actor, upper=1, opposite="ownedAttribute")
Actor.ownedAttribute = association(
    "ownedAttribute", Property, composite=True, opposite="actor"
)
InteractionFragment.enclosingInteraction = association(
    "enclosingInteraction", Interaction, upper=1, opposite="fragment"
)
Interaction.fragment = association(
    "fragment", InteractionFragment, opposite="enclosingInteraction"
)
Constraint.stateInvariant = association(
    "stateInvariant", StateInvariant, upper=1, opposite="invariant"
)
StateInvariant.invariant = association(
    "invariant", Constraint, lower=1, upper=1, composite=True, opposite="stateInvariant"
)
Lifeline.coveredBy = association("coveredBy", InteractionFragment, opposite="covered")
InteractionFragment.covered = association(
    "covered", Lifeline, lower=1, upper=1, opposite="coveredBy"
)
Lifeline.interaction = association(
    "interaction", Interaction, lower=1, upper=1, opposite="lifeline"
)
Interaction.lifeline = association(
    "lifeline", Lifeline, composite=True, opposite="interaction"
)
# 'Message.argument' is a simple attribute
Message.argument = attribute("argument", str)
MessageEnd.sendMessage = association(
    "sendMessage", Message, upper=1, opposite="sendEvent"
)
Message.sendEvent = association(
    "sendEvent", MessageEnd, upper=1, composite=True, opposite="sendMessage"
)
MessageEnd.receiveMessage = association(
    "receiveMessage", Message, upper=1, opposite="receiveEvent"
)
Message.receiveEvent = association(
    "receiveEvent", MessageEnd, upper=1, composite=True, opposite="receiveMessage"
)
Message.interaction = association(
    "interaction", Interaction, lower=1, upper=1, opposite="message"
)
Interaction.message = association(
    "message", Message, composite=True, opposite="interaction"
)
StructuredClassifier.ownedConnector = association(
    "ownedConnector", Connector, composite=True
)
Connector.redefinedConnector = association("redefinedConnector", Connector)
Connector.type = association("type", Association, upper=1)
Connector.end = association("end", ConnectorEnd, lower=2, composite=True)
Connector.contract = association("contract", Behavior)
ConnectorEnd.role = association("role", ConnectableElement, upper=1, opposite="end")
ConnectableElement.end = association("end", ConnectorEnd, opposite="role")
StructuredClassifier.ownedAttribute = association(
    "ownedAttribute", Property, composite=True
)
# 'ObjectNode.upperBound' is a simple attribute
ObjectNode.upperBound = attribute("upperBound", str)
ObjectNode.selection = association("selection", Behavior, upper=1)
# 'JoinNode.joinSpec' is a simple attribute
JoinNode.joinSpec = attribute("joinSpec", str)
StateMachine.region = association(
    "region", Region, lower=1, composite=True, opposite="stateMachine"
)
Region.stateMachine = association(
    "stateMachine", StateMachine, upper=1, opposite="region"
)
Transition.container = association("container", Region, lower=1, upper=1)
Region.subvertex = association(
    "subvertex", Vertex, composite=True, opposite="container"
)
Vertex.container = association("container", Region, upper=1, opposite="subvertex")
Transition.source = association("source", Vertex, lower=1, upper=1, opposite="outgoing")
Vertex.outgoing = association("outgoing", Transition, opposite="source")
Transition.target = association("target", Vertex, lower=1, upper=1, opposite="incoming")
Vertex.incoming = association("incoming", Transition, opposite="target")
ConnectionPointReference.entry = association("entry", Pseudostate)
ConnectionPointReference.exit = association("exit", Pseudostate)
Pseudostate.stateMachine = association("stateMachine", StateMachine, upper=1)
Region.state = association("state", State, upper=1)
Pseudostate.state = association("state", State, upper=1)
ConnectionPointReference.state = association("state", State, upper=1)
State.entry = association("entry", Behavior, upper=1, composite=True)
State.exit = association("exit", Behavior, upper=1, composite=True)
State.doActivity = association("doActivity", Behavior, upper=1, composite=True)
Transition.effect = association("effect", Behavior, upper=1, composite=True)
State.statevariant = association(
    "statevariant", Constraint, upper=1, composite=True, opposite="owningState"
)
Constraint.owningState = association(
    "owningState", State, upper=1, opposite="statevariant"
)
Transition.guard = association("guard", Constraint, upper=1, composite=True)
State.submachine = association("submachine", StateMachine, upper=1)
StateMachine.extendedStateMachine = association(
    "extendedStateMachine", StateMachine, upper=1
)
ConnectorEnd.partWithPort = association("partWithPort", Property, upper=1)
Port.encapsulatedClassifier = association(
    "encapsulatedClassifier", EncapsulatedClassifer, upper=1, opposite="ownedPort"
)
EncapsulatedClassifer.ownedPort = association(
    "ownedPort", Port, composite=True, opposite="encapsulatedClassifier"
)
Element.appliedStereotype = association(
    "appliedStereotype", InstanceSpecification, opposite="extended"
)
InstanceSpecification.extended = association(
    "extended", Element, opposite="appliedStereotype"
)
Node.nestedNode = association("nestedNode", Node, composite=True)
DeploymentTarget.deployment = association("deployment", Deployment, composite=True)
Deployment.deployedArtifact = association("deployedArtifact", DeployedArtifact)
ActivityNode.inPartition = association(
    "inPartition", ActivityPartition, opposite="node"
)
ActivityPartition.node = association("node", ActivityNode, opposite="inPartition")
ActivityPartition.represents = association("represents", Element, upper=1)
ActivityPartition.subpartition = association("subpartition", ActivityPartition)
Association.navigableOwnedEnd = association("navigableOwnedEnd", Property)
AcceptEventAction.result = association("result", OutputPin, composite=True)
UnmarshallAction.result = association("result", OutputPin, composite=True)
AcceptCallAction.returnInformation = association(
    "returnInformation", OutputPin, lower=1, upper=1, composite=True
)
UnmarshallAction.unmarshallType = association(
    "unmarshallType", Classifier, lower=1, upper=1
)
UnmarshallAction.object = association(
    "object", InputPin, lower=1, upper=1, composite=True
)
ReplyAction.replyValue = association("replyValue", InputPin, upper=1, composite=True)
ReplyAction.returnInformation = association(
    "returnInformation", InputPin, lower=1, upper=1, composite=True
)
SendSignalAction.target = association("target", InputPin, composite=True)
Collaboration.collaborationRole = association("collaborationRole", ConnectableElement)
Trigger.event = association("event", Event, lower=1, upper=1)
Signal.ownedAttribute = association("ownedAttribute", Property, composite=True)
Reception.signal = association("signal", Signal, upper=1)
Class.ownedReception = association("ownedReception", Reception, composite=True)
Interface.ownedReception = association("ownedReception", Reception, composite=True)
Action.interaction = association("interaction", Interaction, upper=1, opposite="action")
Interaction.action = association(
    "action", Action, composite=True, opposite="interaction"
)
Message.signature = association("signature", NamedElement, upper=1)
InteractionFragment.generalOrdering = association(
    "generalOrdering", GeneralOrdering, composite=True, opposite="interactionFragment"
)
GeneralOrdering.interactionFragment = association(
    "interactionFragment", InteractionFragment, upper=1, opposite="generalOrdering"
)
ExecutionSpecification.executionOccurrenceSpecification = association(
    "executionOccurrenceSpecification",
    ExecutionOccurrenceSpecification,
    upper=2,
    composite=True,
    opposite="execution",
)
ExecutionOccurrenceSpecification.execution = association(
    "execution",
    ExecutionSpecification,
    lower=1,
    upper=1,
    opposite="executionOccurrenceSpecification",
)
ActionExecutionSpecification.action = association("action", Action, lower=1, upper=1)
BehaviorExecutionSpecification.behavior = association("behavior", Behavior, upper=1)
# 'ChangeEvent.changeExpression' is a simple attribute
ChangeEvent.changeExpression = attribute("changeExpression", str)
Constraint.parameterSet = association(
    "parameterSet", ParameterSet, upper=1, opposite="condition"
)
ParameterSet.condition = association(
    "condition", Constraint, composite=True, opposite="parameterSet"
)
# 47: override MultiplicityElement.lower(MultiplicityElement.lowerValue): attribute[str]
MultiplicityElement.lower = MultiplicityElement.lowerValue

# 50: override MultiplicityElement.upper(MultiplicityElement.upperValue): attribute[str]
MultiplicityElement.upper = MultiplicityElement.upperValue

# 91: override Property.isComposite(Property.aggregation): derived[bool]
Property.isComposite = derived(
    Property, "isComposite", bool, 0, 1, lambda obj: [obj.aggregation == "composite"]
)

# 97: override Property.navigability(Property.opposite, Property.association): derived[Optional[bool]]
# defined in umloverrides.py

RedefinableElement.redefinedElement = derivedunion(
    RedefinableElement,
    "redefinedElement",
    RedefinableElement,
    0,
    "*",
    Property.redefinedProperty,
    Classifier.redefinedClassifier,
    Operation.redefinedOperation,
    Interface.redefinedInterface,
    Behavior.redefinedBehavior,
    Connector.redefinedConnector,
)
Classifier.attribute = derivedunion(
    Classifier,
    "attribute",
    Property,
    0,
    "*",
    Class.ownedAttribute,
    DataType.ownedAttribute,
    Interface.ownedAttribute,
    UseCase.ownedAttribute,
    Actor.ownedAttribute,
    Signal.ownedAttribute,
)
Classifier.feature = derivedunion(
    Classifier,
    "feature",
    Feature,
    0,
    "*",
    Interface.ownedOperation,
    UseCase.extensionPoint,
    DataType.ownedOperation,
    Class.ownedOperation,
    Association.ownedEnd,
    Classifier.attribute,
    StructuredClassifier.ownedConnector,
    Class.ownedReception,
    Interface.ownedReception,
)
Feature.featuringClassifier = derivedunion(
    Feature,
    "featuringClassifier",
    Classifier,
    1,
    "*",
    Property.class_,
    Property.owningAssociation,
    Operation.class_,
    Operation.datatype,
    Property.datatype,
    Operation.interface_,
)
# 88: override Property.opposite(Property.association, Association.memberEnd): relation_one[Optional[Property]]
# defined in umloverrides.py

BehavioralFeature.parameter = derivedunion(
    BehavioralFeature,
    "parameter",
    Parameter,
    0,
    "*",
    BehavioralFeature.returnResult,
    BehavioralFeature.formalParameter,
)
Action.output = derivedunion(Action, "output", OutputPin, 0, "*")
RedefinableElement.redefinitionContext = derivedunion(
    RedefinableElement,
    "redefinitionContext",
    Classifier,
    0,
    "*",
    Operation.class_,
    Property.classifier,
    Operation.datatype,
    Port.encapsulatedClassifier,
)
NamedElement.namespace = derivedunion(
    NamedElement,
    "namespace",
    Namespace,
    0,
    1,
    Parameter.ownerReturnParam,
    Property.interface_,
    Property.class_,
    Property.owningAssociation,
    Operation.class_,
    EnumerationLiteral.enumeration,
    Diagram.package,
    Operation.datatype,
    Type.package,
    Property.datatype,
    Operation.interface_,
    Package.package,
    Parameter.ownerFormalParam,
    Property.useCase,
    Property.actor,
    InteractionFragment.enclosingInteraction,
    Lifeline.interaction,
    Message.interaction,
    Region.stateMachine,
    Transition.container,
    Vertex.container,
    Pseudostate.stateMachine,
    Region.state,
    ConnectionPointReference.state,
)
Namespace.ownedMember = derivedunion(
    Namespace,
    "ownedMember",
    NamedElement,
    0,
    "*",
    Interface.ownedOperation,
    Enumeration.literal,
    Package.ownedDiagram,
    Namespace.ownedRule,
    UseCase.extensionPoint,
    DataType.ownedOperation,
    Operation.precondition,
    BehavioralFeature.returnResult,
    Profile.ownedStereotype,
    Class.nestedClassifier,
    Class.ownedAttribute,
    BehavioralFeature.formalParameter,
    Classifier.ownedUseCase,
    DataType.ownedAttribute,
    Class.ownedOperation,
    Operation.postcondition,
    Association.ownedEnd,
    Package.ownedClassifier,
    Interface.ownedAttribute,
    Operation.bodyCondition,
    Extend.constraint,
    Package.nestedPackage,
    BehavioredClassifier.ownedBehavior,
    UseCase.ownedAttribute,
    Actor.ownedAttribute,
    Interaction.fragment,
    Interaction.lifeline,
    Interaction.message,
    StateMachine.region,
    Region.subvertex,
    Node.nestedNode,
    Signal.ownedAttribute,
    Class.ownedReception,
    Interface.ownedReception,
)
# 79: override Classifier.general(Generalization.general): derived[Classifier]
Classifier.general = derived(
    Classifier,
    "general",
    Classifier,
    0,
    "*",
    lambda self: [g.general for g in self.generalization],
)

# 53: override Association.endType(Association.memberEnd, Property.type): derived[Type]

# References the classifiers that are used as types of the ends of the
# association.

Association.endType = derived(
    Association,
    "endType",
    Type,
    0,
    "*",
    lambda self: [end.type for end in self.memberEnd if end],
)


# 94: override Constraint.context: derivedunion[Namespace]
Constraint.context = derivedunion(Constraint, "context", Namespace, 0, 1)

# 100: override Operation.type: derivedunion[DataType]
Operation.type = derivedunion(Operation, "type", DataType, 0, 1)

# 73: override Extension.metaclass(Extension.ownedEnd, Association.memberEnd): property
# defined in umloverrides.py

# 61: override Class.extension(Extension.metaclass): property
# See https://www.omg.org/spec/UML/2.5/PDF, section 11.8.3.6, page 219
# It defines `Extension.allInstances()`, which basically means we have to query the element factory.

# TODO: use those as soon as Extension.metaclass can be used.
# Class.extension = derived(Class, 'extension', Extension, 0, '*', class_extension, Extension.metaclass)

Class.extension = property(
    lambda self: self.model.lselect(
        lambda e: e.isKindOf(Extension) and self is e.metaclass
    ),
    doc="""References the Extensions that specify additional properties of the
metaclass. The property is derived from the extensions whose memberEnds
are typed by the Class.""",
)

DirectedRelationship.target = derivedunion(
    DirectedRelationship,
    "target",
    Element,
    1,
    "*",
    PackageImport.importedPackage,
    PackageMerge.mergedPackage,
    Generalization.general,
    Include.addition,
    Extend.extendedCase,
    Realization.realizingClassifier,
    ElementImport.importedElement,
    Substitution.contract,
)
DirectedRelationship.source = derivedunion(
    DirectedRelationship,
    "source",
    Element,
    1,
    "*",
    Extend.extension,
    Realization.abstraction,
    Substitution.substitutingClassifier,
    Include.includingCase,
    ElementImport.importingNamespace,
    Generalization.specific,
    PackageImport.importingNamespace,
    PackageMerge.mergingPackage,
)
Action.context_ = derivedunion(Action, "context_", Classifier, 0, 1)
Relationship.relatedElement = derivedunion(
    Relationship,
    "relatedElement",
    Element,
    1,
    "*",
    DirectedRelationship.target,
    DirectedRelationship.source,
)
ActivityGroup.superGroup = derivedunion(
    ActivityGroup, "superGroup", ActivityGroup, 0, 1
)
ActivityGroup.subgroup = derivedunion(
    ActivityGroup, "subgroup", ActivityGroup, 0, "*", ActivityPartition.subpartition
)
# 76: override Classifier.inheritedMember: derivedunion[NamedElement]
Classifier.inheritedMember = derivedunion(
    Classifier, "inheritedMember", NamedElement, 0, "*"
)

StructuredClassifier.role = derivedunion(
    StructuredClassifier,
    "role",
    ConnectableElement,
    0,
    "*",
    StructuredClassifier.ownedAttribute,
    Collaboration.collaborationRole,
)
Namespace.member = derivedunion(
    Namespace,
    "member",
    NamedElement,
    0,
    "*",
    BehavioralFeature.parameter,
    Namespace.ownedMember,
    Association.memberEnd,
    Classifier.inheritedMember,
    StructuredClassifier.role,
)
# 112: override Component.required: property
# defined in umloverrides.py

# 85: override Namespace.importedMember: derivedunion[PackageableElement]
Namespace.importedMember = derivedunion(
    Namespace, "importedMember", PackageableElement, 0, "*"
)

Action.input = derivedunion(Action, "input", InputPin, 0, "*", SendSignalAction.target)
# 109: override Component.provided: property
# defined in umloverrides.py

Element.owner = derivedunion(
    Element,
    "owner",
    Element,
    0,
    1,
    Slot.owningInstance,
    Realization.abstraction,
    ElementImport.importingNamespace,
    Generalization.specific,
    ActivityEdge.activity,
    ActivityGroup.superGroup,
    ActivityGroup.activity,
    PackageImport.importingNamespace,
    PackageMerge.mergingPackage,
    NamedElement.namespace,
    Constraint.stateInvariant,
    Pseudostate.state,
    Action.interaction,
    GeneralOrdering.interactionFragment,
    Constraint.parameterSet,
)
Element.ownedElement = derivedunion(
    Element,
    "ownedElement",
    Element,
    0,
    "*",
    Artifact.manifestation,
    Element.ownedComment,
    Action.input,
    Classifier.generalization,
    Namespace.ownedMember,
    Namespace.elementImport,
    Activity.group,
    Component.realization,
    Namespace.packageImport,
    Package.packageExtension,
    Substitution.contract,
    ActivityGroup.subgroup,
    Activity.edge,
    Activity.node,
    Action.output,
    StateInvariant.invariant,
    Connector.end,
    State.entry,
    State.exit,
    State.doActivity,
    Transition.effect,
    State.statevariant,
    Transition.guard,
    DeploymentTarget.deployment,
    Interaction.action,
    InteractionFragment.generalOrdering,
    ParameterSet.condition,
)
ConnectorEnd.definingEnd = derivedunion(ConnectorEnd, "definingEnd", Property, 0, 1)
# 118: override StructuredClassifier.part: property
StructuredClassifier.part = property(
    lambda self: tuple(a for a in self.ownedAttribute if a.isComposite),
    doc="""
    Properties owned by a classifier by composition.
""",
)

# 123: override ExecutionSpecification.start(ExecutionSpecification.executionOccurrenceSpecification): relation_one[ExecutionOccurrenceSpecification]
ExecutionSpecification.start = derived(
    ExecutionSpecification,
    "start",
    OccurrenceSpecification,
    0,
    1,
    lambda obj: [
        eos for i, eos in enumerate(obj.executionOccurrenceSpecification) if i == 0
    ],
)

# 127: override ExecutionSpecification.finish(ExecutionSpecification.executionOccurrenceSpecification): relation_one[ExecutionOccurrenceSpecification]
ExecutionSpecification.finish = derived(
    ExecutionSpecification,
    "finish",
    OccurrenceSpecification,
    0,
    1,
    lambda obj: [
        eos for i, eos in enumerate(obj.executionOccurrenceSpecification) if i == 1
    ],
)

# 82: override Class.superClass: derived[Classifier]
Class.superClass = Classifier.general

ExtensionEnd.type = redefine(ExtensionEnd, "type", Stereotype, 1, Property.type)
ActivityNode.redefinedElement = redefine(
    ActivityNode,
    "redefinedElement",
    ActivityNode,
    "*",
    RedefinableElement.redefinedElement,
)
Implementation.contract = redefine(
    Implementation, "contract", Interface, "*", Dependency.supplier
)
BehavioredClassifier.implementation = redefine(
    BehavioredClassifier,
    "implementation",
    Implementation,
    "*",
    NamedElement.clientDependency,
)
Implementation.implementatingClassifier = redefine(
    Implementation,
    "implementatingClassifier",
    BehavioredClassifier,
    "*",
    Dependency.client,
)
Parameter.operation = redefine(
    Parameter, "operation", Operation, 1, Parameter.ownerFormalParam
)
Operation.formalParameter = redefine(
    Operation, "formalParameter", Parameter, "*", BehavioralFeature.formalParameter
)
ActivityEdge.redefinedElement = redefine(
    ActivityEdge,
    "redefinedElement",
    ActivityEdge,
    "*",
    RedefinableElement.redefinedElement,
)
Package.ownedMember = redefine(
    Package, "ownedMember", PackageableElement, "*", Namespace.ownedMember
)
Component.ownedMember = redefine(
    Component, "ownedMember", PackageableElement, "*", Namespace.ownedMember
)
Transition.redefinitionContext = redefine(
    Transition,
    "redefinitionContext",
    Classifier,
    "*",
    RedefinableElement.redefinitionContext,
)
Region.extendedRegion = redefine(
    Region, "extendedRegion", Region, "*", RedefinableElement.redefinedElement
)
State.redefinedState = redefine(
    State, "redefinedState", State, "*", RedefinableElement.redefinedElement
)
Transition.redefinedTransition = redefine(
    Transition,
    "redefinedTransition",
    Transition,
    "*",
    RedefinableElement.redefinedElement,
)
StateInvariant.covered = redefine(
    StateInvariant, "covered", Lifeline, 1, InteractionFragment.covered
)
OccurrenceSpecification.covered = redefine(
    OccurrenceSpecification, "covered", Lifeline, 1, InteractionFragment.covered
)
# 103: override Lifeline.parse: Callable[[Lifeline, str], None]
# defined in umloverrides.py

# 106: override Lifeline.render: Callable[[Lifeline], str]
# defined in umloverrides.py



import pytest

from gaphor.diagram.general import Box
from gaphor.plugins.diagramexport import DiagramExport


@pytest.fixture
def diagram_export():
    return DiagramExport(diagrams=None)


@pytest.fixture
def diagram_with_box(diagram):
    diagram.create(Box)
    return diagram


def test_export_to_svg(diagram_export, diagram_with_box, tmp_path):
    f = tmp_path / "test.svg"

    diagram_export.save_svg(f, diagram_with_box)
    content = f.read_text()

    assert "<svg" in content


def test_export_to_png(diagram_export, diagram_with_box, tmp_path):
    f = tmp_path / "test.png"

    diagram_export.save_png(f, diagram_with_box)
    content = f.read_bytes()

    assert b"PNG" in content


def test_export_to_pdf(diagram_export, diagram_with_box, tmp_path):
    f = tmp_path / "test.pdf"

    diagram_export.save_pdf(f, diagram_with_box)
    content = f.read_bytes()

    assert b"%PDF" in content



"""Unit tests for transactions in Gaphor."""
import pytest

from gaphor.core import event_handler
from gaphor.core.eventmanager import EventManager
from gaphor.event import TransactionBegin, TransactionCommit, TransactionRollback
from gaphor.transaction import Transaction, TransactionError

begins = []
commits = []
rollbacks = []


@event_handler(TransactionBegin)
def handle_begins(event):
    """Store TransactionBegin events in begins."""
    begins.append(event)


@event_handler(TransactionCommit)
def handle_commits(event):
    """Store TransactionCommit events in commits."""
    commits.append(event)


@event_handler(TransactionRollback)
def handle_rollback(event):
    """Store TransactionRollback events in rollbacks."""
    rollbacks.append(event)


@pytest.fixture
def event_manager():
    event_manager = EventManager()

    event_manager.subscribe(handle_begins)
    event_manager.subscribe(handle_commits)
    event_manager.subscribe(handle_rollback)

    del begins[:]
    del commits[:]
    del rollbacks[:]

    yield event_manager

    event_manager.unsubscribe(handle_begins)
    event_manager.unsubscribe(handle_commits)
    event_manager.unsubscribe(handle_rollback)


def test_transaction_commit(event_manager):
    """Test committing a transaction."""

    tx = Transaction(event_manager)

    assert tx._stack, "Transaction has no stack"
    assert len(begins) == 1, "Incorrect number of TrasactionBegin events"
    assert len(commits) == 0, "Incorrect number of TransactionCommit events"
    assert len(rollbacks) == 0, "Incorrect number of TransactionRollback events"

    tx.commit()

    assert len(begins) == 1, "Incorrect number of TrasactionBegin events"
    assert len(commits) == 1, "Incorrect number of TransactionCommit events"
    assert len(rollbacks) == 0, "Incorrect number of TransactionRollback events"
    assert not tx._stack, "Transaction stack is not empty"

    with pytest.raises(TransactionError):
        tx.commit()


def test_transaction_rollback(event_manager):
    """Test rolling back a transaction."""

    tx = Transaction(event_manager)

    assert tx._stack, "Transaction has no stack"
    assert len(begins) == 1, "Incorrect number of TrasactionBegin events"
    assert len(commits) == 0, "Incorrect number of TransactionCommit events"
    assert len(rollbacks) == 0, "Incorrect number of TransactionRollback events"

    tx.rollback()

    assert len(begins) == 1, "Incorrect number of TrasactionBegin events"
    assert len(commits) == 0, "Incorrect number of TransactionCommit events"
    assert len(rollbacks) == 1, "Incorrect number of TransactionRollback events"

    assert not tx._stack, "Transaction stack is not empty"


def test_transaction_commit_after_rollback(event_manager):
    """Test committing one transaction after rolling back another
    transaction."""

    tx1 = Transaction(event_manager)
    tx2 = Transaction(event_manager)

    tx2.rollback()
    tx1.commit()

    assert len(begins) == 1, "Incorrect number of TrasactionBegin events"
    assert len(commits) == 0, "Incorrect number of TransactionCommit events"
    assert len(rollbacks) == 1, "Incorrect number of TransactionRollback events"


def test_transaction_rollback_after_commit(event_manager):
    """Test committing one transaction after rolling back another
    transaction."""

    tx1 = Transaction(event_manager)
    tx2 = Transaction(event_manager)

    tx2.commit()
    tx1.rollback()

    assert len(begins) == 1, "Incorrect number of TrasactionBegin events"
    assert len(commits) == 0, "Incorrect number of TransactionCommit events"
    assert len(rollbacks) == 1, "Incorrect number of TransactionRollback events"


def test_transaction_stack(event_manager):
    """Test the transaction stack."""

    tx1 = Transaction(event_manager)
    tx2 = Transaction(event_manager)

    with pytest.raises(TransactionError):
        tx1.commit()

    tx2.rollback()
    tx1.rollback()


def test_transaction_context(event_manager):
    """Test the transaction context manager."""

    with Transaction(event_manager) as tx:

        assert isinstance(tx, Transaction), "Context is not a Transaction instance"
        assert Transaction._stack, "Transaction instance has no stack inside a context"

    assert not Transaction._stack, "Transaction stack should be empty"


def test_transaction_context_error(event_manager):
    """Test the transaction context manager with errors."""

    with pytest.raises(TypeError, match="transaction error"):
        with Transaction(event_manager):
            raise TypeError("transaction error")



import itertools
from typing import List, NamedTuple

from gaphor.diagram.copypaste import (
    ElementCopy,
    NamedElementCopy,
    copy,
    copy_element,
    copy_named_element,
    paste,
    paste_element,
    paste_named_element,
)
from gaphor.UML import Association, Class, Interface


class ClassCopy(NamedTuple):
    element_copy: NamedElementCopy
    owned_attributes: List[ElementCopy]
    owned_parameters: List[ElementCopy]
    owned_operations: List[ElementCopy]


@copy.register(Class)
@copy.register(Interface)
def copy_class(element):
    return ClassCopy(
        element_copy=copy_named_element(element),
        owned_attributes=[
            copy_element(attr)
            for attr in element.ownedAttribute
            if not attr.association
        ],
        owned_parameters=[
            copy_element(oper)
            for oper in itertools.chain(
                element.ownedOperation[:].formalParameter,
                element.ownedOperation[:].returnResult,
            )
        ],
        owned_operations=[copy_element(oper) for oper in element.ownedOperation],
    )


@paste.register
def paste_class(copy_data: ClassCopy, diagram, lookup):
    for attr in itertools.chain(
        copy_data.owned_attributes,
        copy_data.owned_parameters,
        copy_data.owned_operations,
    ):
        paste_element(attr, diagram, lookup)
    return paste_named_element(copy_data.element_copy, diagram, lookup)


class AssociationCopy(NamedTuple):
    element_copy: NamedElementCopy
    member_ends: List[ElementCopy]


@copy.register
def copy_association(element: Association):
    return AssociationCopy(
        element_copy=copy_named_element(element),
        member_ends=[copy_element(end) for end in element.memberEnd],
    )


@paste.register
def paste_association(copy_data: AssociationCopy, diagram, lookup):
    for member_end in copy_data.member_ends:
        paste_element(member_end, diagram, lookup)
    return paste_named_element(copy_data.element_copy, diagram, lookup)



# Here, order matters
import gaphor.UML.umloverrides
from gaphor.UML import modelfactory as model
from gaphor.UML.uml import *
from gaphor.UML.umlfmt import format
from gaphor.UML.umllex import parse



from gaphor import UML
from gaphor.core.modeling.diagram import StyledItem
from gaphor.UML.classes import ClassItem


def test_attribute_of_item(diagram):
    classitem = diagram.create(ClassItem)

    node = StyledItem(classitem)

    assert node.attribute("show_attributes") == "true"


def test_nonexistant_attribute_of_item(diagram):
    classitem = diagram.create(ClassItem)

    node = StyledItem(classitem)

    assert node.attribute("foobar") == ""


def test_should_get_type_of_element_attribute(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    classitem = diagram.create(ClassItem, subject=class_)

    node = StyledItem(classitem)

    assert node.attribute("subject") == "class"


def test_should_read_attribute_of_subject(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    classitem = diagram.create(ClassItem, subject=class_)

    class_.name = "myname"
    node = StyledItem(classitem)

    assert node.attribute("name") == "myname"


# test attributes that can be any of a list
def test_nested_attribute_of_subject(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    attr = element_factory.create(UML.Property)
    class_.ownedAttribute = attr
    classitem = diagram.create(ClassItem, subject=class_)

    attr.name = "myname"
    attr.isService = 1
    node = StyledItem(classitem)

    assert node.attribute("attribute") == "property"
    assert node.attribute("attribute.name") == "myname"
    assert node.attribute("attribute.isBehavior") == ""
    assert node.attribute("attribute.isService") == "true"
    assert node.attribute("attribute.notAnAttribute") == ""


def test_multiple_nested_attribute_of_subject(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    attr1 = element_factory.create(UML.Property)
    attr2 = element_factory.create(UML.Property)
    class_.ownedAttribute = attr1
    class_.ownedAttribute = attr2
    classitem = diagram.create(ClassItem, subject=class_)

    attr1.name = "first"
    attr2.name = "second"
    node = StyledItem(classitem)

    assert node.attribute("attribute") == "property property"
    assert node.attribute("attribute.name") in ("first second", "second first")
    assert node.attribute("attribute.notAnAttribute") == ""


def test_mixed_case_attributes(diagram, element_factory):
    class_ = element_factory.create(UML.Class)
    attr1 = element_factory.create(UML.Property)
    class_.ownedAttribute = attr1
    classitem = diagram.create(ClassItem, subject=class_)

    attr1.name = "first"
    node = StyledItem(classitem)

    assert node.attribute("ownedattribute") == "property"
    assert node.attribute("ownedattribute.name") == "first"



import os
from pathlib import Path

from gi.repository import Gio, GLib, Gtk

from gaphor.abc import Service
from gaphor.core import event_handler, gettext
from gaphor.ui import APPLICATION_ID
from gaphor.ui.event import FileLoaded, FileSaved

HOME = str(Path.home())


class RecentFiles(Service):
    def __init__(self, event_manager, recent_manager=None):
        self.event_manager = event_manager
        self.recent_manager = recent_manager or Gtk.RecentManager.get_default()

        self.event_manager.subscribe(self._on_filename_changed)

    def shutdown(self):
        self.event_manager.unsubscribe(self._on_filename_changed)

    @event_handler(FileLoaded, FileSaved)
    def _on_filename_changed(self, event):
        filename = event.filename
        if not filename:
            return
        uri = GLib.filename_to_uri(os.path.abspath(filename))
        meta = Gtk.RecentData()
        meta.app_name = APPLICATION_ID
        meta.app_exec = f"{GLib.get_prgname()} %u"
        meta.mime_type = "application/x-gaphor"
        self.recent_manager.add_full(uri, meta)


class RecentFilesMenu(Gio.Menu):
    def __init__(self, recent_manager):
        super().__init__()

        self._on_recent_manager_changed(recent_manager)
        # TODO: should unregister if the window is closed.
        self._changed_id = recent_manager.connect(
            "changed", self._on_recent_manager_changed
        )

    def _on_recent_manager_changed(self, recent_manager):
        self.remove_all()
        for item in recent_manager.get_items():
            if APPLICATION_ID in item.get_applications():
                menu_item = Gio.MenuItem.new(
                    item.get_uri_display().replace(HOME, "~"), "app.file-open-recent"
                )
                filename, _host = GLib.filename_from_uri(item.get_uri())
                menu_item.set_attribute_value(
                    "target", GLib.Variant.new_string(filename)
                )
                self.append_item(menu_item)
                if self.get_n_items() > 9:
                    break
        if self.get_n_items() == 0:
            self.append_item(
                Gio.MenuItem.new(gettext("No recently opened models"), None)
            )



"""
The Sanitize module is dedicated to adapters (stuff) that keeps
the model clean and in sync with diagrams.
"""


from gaphor import UML
from gaphor.abc import Service
from gaphor.core import event_handler
from gaphor.core.modeling import Presentation
from gaphor.core.modeling.event import AssociationDeleted, AssociationSet
from gaphor.diagram.general import CommentLineItem


class SanitizerService(Service):
    """
    Does some background cleanup jobs, such as removing elements from the
    model that have no presentations (and should have some).
    """

    def __init__(self, event_manager):
        self.event_manager = event_manager

        event_manager.subscribe(self._unlink_on_presentation_delete)
        event_manager.subscribe(self.update_annotated_element_link)
        event_manager.subscribe(self._unlink_on_stereotype_delete)
        event_manager.subscribe(self._unlink_on_extension_delete)
        event_manager.subscribe(self._disconnect_extension_end)

    def shutdown(self):
        event_manager = self.event_manager
        event_manager.unsubscribe(self._unlink_on_presentation_delete)
        event_manager.unsubscribe(self.update_annotated_element_link)
        event_manager.unsubscribe(self._unlink_on_stereotype_delete)
        event_manager.unsubscribe(self._unlink_on_extension_delete)
        event_manager.unsubscribe(self._disconnect_extension_end)

    @event_handler(AssociationDeleted)
    def _unlink_on_presentation_delete(self, event):
        """
        Unlink the model element if no more presentations link to the `item`'s
        subject or the deleted item was the only item currently linked.
        """
        if event.property is UML.Element.presentation:
            old_presentation = event.old_value
            if old_presentation and not event.element.presentation:
                event.element.unlink()

    @event_handler(AssociationSet)
    def update_annotated_element_link(self, event):
        """
        Link comment and element if a comment line is present, but comment
        and element subject are not connected yet.
        """
        if event.property is not Presentation.subject:  # type: ignore[misc]
            return

        element: Presentation = event.element
        subject = event.new_value
        if not (element.canvas and subject):
            return

        for cinfo in element.canvas.get_connections(connected=element):
            comment_line = cinfo.item
            if not isinstance(comment_line, CommentLineItem):
                continue
            opposite = comment_line.opposite(cinfo.handle)
            opposite_cinfo = element.canvas.get_connection(opposite)
            if not opposite_cinfo:
                continue
            comment_item = opposite_cinfo.connected
            comment = comment_item.subject
            comment.annotatedElement = subject

    def perform_unlink_for_instances(self, st, meta):
        inst = UML.model.find_instances(st)

        for i in list(inst):
            for e in i.extended:
                if not meta or isinstance(e, meta):
                    i.unlink()

    @event_handler(AssociationDeleted)
    def _unlink_on_extension_delete(self, event):
        """
        Remove applied stereotypes when extension is deleted.
        """
        if (
            isinstance(event.element, UML.Extension)
            and event.property is UML.Association.memberEnd
            and event.element.memberEnd
        ):
            p = event.element.memberEnd[0]
            ext = event.old_value
            if isinstance(p, UML.ExtensionEnd):
                p, ext = ext, p
            st = ext.type
            meta = p.type and getattr(UML, p.type.name, None)
            if st:
                self.perform_unlink_for_instances(st, meta)

    @event_handler(AssociationSet)
    def _disconnect_extension_end(self, event):
        if event.property is UML.ExtensionEnd.type and event.old_value:
            ext = event.element
            p = ext.opposite
            if not p:
                return
            st = event.old_value
            if st:
                meta = getattr(UML, p.type.name, None)
                self.perform_unlink_for_instances(st, meta)

    @event_handler(AssociationDeleted)
    def _unlink_on_stereotype_delete(self, event):
        """
        Remove applied stereotypes when stereotype is deleted.
        """
        if event.property is UML.InstanceSpecification.classifier and isinstance(
            event.old_value, UML.Stereotype
        ):
            event.element.unlink()



"""Internationalization (i18n) support for Gaphor.

Translate text in to your native language using the gettext() function.

"""
__all__ = ["gettext"]

import gettext as _gettext
import importlib.resources
import logging

log = logging.getLogger(__name__)

try:

    with importlib.resources.path("gaphor", "locale") as path:
        translate = _gettext.translation("gaphor", localedir=str(path), fallback=True)
        gettext = translate.gettext

except OSError as e:
    log.info(f"No translations were found: {e}")

    def gettext(s):
        return s



"""
Grouping functionality allows to nest one item within another item (parent
item). This is useful in several use cases

- artifact deployed within a node
- a class within a package or a component
- composite structures (i.e. component within a node)

The grouping adapters has to implement three methods, see `AbstractGroup`
class.

It is important to note, that grouping adapters can be queried before
instance of an item to be grouped is created. This happens when item
is about to be created. Therefore `AbstractGroup.can_contain` has
to be aware that `AbstractGroup.item` can be null.
"""

from __future__ import annotations

import abc
from typing import Type

from generic.multidispatch import FunctionDispatcher, multidispatch

from gaphor.core.modeling import Presentation


# TODO: I think this should have been called Namespacing or something similar,
#   since that's the modeling concept.
class AbstractGroup(metaclass=abc.ABCMeta):
    """
    Base class for grouping model elements, i.e.
    interactions contain lifelines and components contain classes objects.

    :param parent: Parent item, which groups other items.
    :type parent: Presentation
    :param item: Item to be grouped.
    :type item: Presentation
    """

    def __init__(self, parent: Presentation, item: Presentation) -> None:
        self.parent = parent
        self.item = item

    def can_contain(self) -> bool:
        """
        Determine if parent can contain item.
        """
        return True

    @abc.abstractmethod
    def group(self) -> None:
        """
        Perform grouping of items.
        """

    @abc.abstractmethod
    def ungroup(self) -> None:
        """
        Perform ungrouping of items.
        """


# Work around issue https://github.com/python/mypy/issues/3135 (Class decorators are not type checked)
# This definition, along with the the ignore below, seems to fix the behaviour for mypy at least.

# @multidispatch(object, object)
class NoGrouping(AbstractGroup):
    def can_contain(self) -> bool:
        return False

    def group(self) -> None:
        pass

    def ungroup(self) -> None:
        pass


Group: FunctionDispatcher[Type[AbstractGroup]] = multidispatch(object, object)(
    NoGrouping
)

# Until we can deal with types (esp. typing.Any) we use this as a workaround:
Group.register(None, object)(NoGrouping)



"""
Message connection adapter tests.
"""

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.UML.interactions.executionspecification import ExecutionSpecificationItem
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.interactions.message import MessageItem


def test_head_glue(diagram):
    """Test message head glue
    """
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    # get head port
    port = ll.ports()[0]
    glued = allow(msg, msg.head, ll, port)
    assert glued


def test_invisible_lifetime_glue(diagram):
    """Test message to invisible lifetime glue
    """
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    glued = allow(msg, msg.head, ll, ll.lifetime.port)

    assert not ll.lifetime.visible
    assert not glued


def test_visible_lifetime_glue(diagram):
    """Test message to visible lifetime glue
    """
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    ll.lifetime.visible = True

    glued = allow(msg, msg.head, ll, ll.lifetime.port)
    assert glued


def test_lost_message_connection(diagram, element_factory):
    """Test lost message connection
    """
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # If one side is connected a "lost" message is created
    assert msg.subject is not None
    assert msg.subject.messageKind == "lost"

    assert len(messages) == 1
    assert len(occurrences) == 1
    assert messages[0] is msg.subject
    assert occurrences[0] is msg.subject.sendEvent


def test_found_message_connection(diagram, element_factory):
    """Test found message connection
    """
    ll = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.tail, ll)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # If one side is connected a "found" message is created
    assert msg.subject is not None
    assert msg.subject.messageKind == "found"

    assert len(messages) == 1
    assert len(occurrences) == 1
    assert messages[0] is msg.subject
    assert occurrences[0] is msg.subject.receiveEvent


def test_complete_message_connection(diagram, element_factory):
    """Test complete message connection
    """
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll1)
    connect(msg, msg.tail, ll2)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # two sides are connected - "complete" message is created
    assert msg.subject is not None
    assert msg.subject.messageKind == "complete"

    assert len(messages) == 1
    assert len(occurrences) == 2
    assert messages[0] is msg.subject
    assert msg.subject.sendEvent in occurrences, f"{occurrences}"
    assert msg.subject.receiveEvent in occurrences, f"{occurrences}"


def test_lifetime_connection(diagram):
    """Test messages' lifetimes connection
    """
    msg = diagram.create(MessageItem)
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)

    # make lifelines to be in sequence diagram mode
    ll1.lifetime.visible = True
    ll2.lifetime.visible = True
    assert ll1.lifetime.visible and ll2.lifetime.visible

    # connect lifetimes with messages message to lifeline's head
    connect(msg, msg.head, ll1, ll1.lifetime.port)
    connect(msg, msg.tail, ll2, ll2.lifetime.port)

    assert msg.subject is not None
    assert msg.subject.messageKind == "complete"


def test_disconnection(diagram):
    """Test message disconnection
    """
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll1)
    connect(msg, msg.tail, ll2)

    # one side disconnection
    disconnect(msg, msg.head)
    assert msg.subject is not None, f"{msg.subject}"

    # 2nd side disconnection
    disconnect(msg, msg.tail)
    assert msg.subject is None, f"{msg.subject}"


def test_lifetime_connectivity_on_head(diagram, element_factory):
    """Test lifeline's lifetime connectivity change on head connection
    """
    ll = diagram.create(LifelineItem, subject=element_factory.create(UML.Lifeline))
    msg = diagram.create(MessageItem)

    # connect message to lifeline's head, lifeline's lifetime
    # visibility and connectivity should change
    connect(msg, msg.head, ll)
    assert not ll.lifetime.visible
    assert not ll.lifetime.connectable
    assert ll.lifetime.MIN_LENGTH == ll.lifetime.min_length

    # ... and disconnection
    disconnect(msg, msg.head)
    assert ll.lifetime.connectable
    assert ll.lifetime.MIN_LENGTH == ll.lifetime.min_length


def test_lifetime_connectivity_on_lifetime(diagram, element_factory):
    """Test lifeline's lifetime connectivity change on lifetime connection
    """
    ll = diagram.create(LifelineItem, subject=element_factory.create(UML.Lifeline))
    msg = diagram.create(MessageItem)

    ll.lifetime.visible = True

    # connect message to lifeline's lifetime, lifeline's lifetime
    # visibility and connectivity should be unchanged
    connect(msg, msg.head, ll, ll.lifetime.port)
    assert ll.lifetime.connectable
    assert ll.lifetime.MIN_LENGTH_VISIBLE == ll.lifetime.min_length

    # ... and disconnection
    disconnect(msg, msg.head)
    assert ll.lifetime.connectable
    assert ll.lifetime.visible
    assert ll.lifetime.MIN_LENGTH == ll.lifetime.min_length


def test_message_glue_cd(diagram):
    """Test gluing message on communication diagram."""

    lifeline1 = diagram.create(LifelineItem)
    lifeline2 = diagram.create(LifelineItem)
    message = diagram.create(MessageItem)

    # make second lifeline to be in sequence diagram mode
    lifeline2.lifetime.visible = True

    # connect head of message to lifeline's head
    connect(message, message.head, lifeline1)

    glued = allow(message, message.tail, lifeline2, lifeline2.lifetime.port)
    # no connection possible as 2nd lifeline is in sequence diagram
    # mode
    assert not glued


def test_message_glue_sd(diagram):
    """Test gluing message on sequence diagram."""

    msg = diagram.create(MessageItem)
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)

    # 1st lifeline - communication diagram
    # 2nd lifeline - sequence diagram
    ll2.lifetime.visible = True

    # connect lifetime of message to lifeline's lifetime
    connect(msg, msg.head, ll1, ll1.lifetime.port)

    glued = allow(msg, msg.tail, ll2)
    # no connection possible as 2nd lifeline is in communication
    # diagram mode
    assert not glued


def test_messages_disconnect_cd(diagram, element_factory):
    """Test disconnecting messages on communication diagram
    """
    ll1 = diagram.create(LifelineItem)
    ll2 = diagram.create(LifelineItem)
    msg = diagram.create(MessageItem)

    connect(msg, msg.head, ll1)
    connect(msg, msg.tail, ll2)

    subject = msg.subject

    assert subject.sendEvent
    assert subject.receiveEvent

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    # verify integrity of messages
    assert len(messages) == 1
    assert len(occurrences) == 2


def test_message_connect_to_execution_specification(diagram, element_factory):
    """Test gluing message on sequence diagram."""

    lifeline = diagram.create(
        LifelineItem, subject=element_factory.create(UML.Lifeline)
    )
    exec_spec = diagram.create(ExecutionSpecificationItem)
    message = diagram.create(MessageItem)
    connect(exec_spec, exec_spec.handles()[0], lifeline, lifeline.lifetime.port)

    connect(message, message.head, exec_spec, exec_spec.ports()[0])

    assert message.subject
    assert message.subject.sendEvent.covered is lifeline.subject


def test_message_disconnect_from_execution_specification(diagram, element_factory):
    """Test gluing message on sequence diagram."""

    lifeline = diagram.create(
        LifelineItem, subject=element_factory.create(UML.Lifeline)
    )
    exec_spec = diagram.create(ExecutionSpecificationItem)
    message = diagram.create(MessageItem)
    connect(exec_spec, exec_spec.handles()[0], lifeline, lifeline.lifetime.port)
    connect(message, message.head, exec_spec, exec_spec.ports()[0])

    disconnect(message, message.head)

    messages = element_factory.lselect(UML.Message)
    occurrences = element_factory.lselect(
        lambda e: e.isKindOf(UML.MessageOccurrenceSpecification)
    )

    assert not message.subject
    assert not len(messages)
    assert not len(occurrences)



import pytest

from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.SysML import sysml
from gaphor.SysML.requirements.relationships import (
    DeriveReqtItem,
    RefineItem,
    SatisfyItem,
    TraceItem,
    VerifyItem,
)
from gaphor.SysML.requirements.requirement import RequirementItem


@pytest.mark.parametrize(
    "item_class", [DeriveReqtItem, RefineItem, SatisfyItem, TraceItem, VerifyItem]
)
def test_relation_allow_connect_disconnect_cycle(diagram, element_factory, item_class):
    req1 = element_factory.create(sysml.Requirement)
    req2 = element_factory.create(sysml.Requirement)

    req_item1 = diagram.create(RequirementItem, subject=req1)
    req_item2 = diagram.create(RequirementItem, subject=req2)

    relation = diagram.create(item_class)

    assert allow(relation, relation.handles()[0], req_item1)
    assert allow(relation, relation.handles()[1], req_item2)

    connect(relation, relation.handles()[0], req_item1)
    connect(relation, relation.handles()[1], req_item2)

    assert relation.subject
    assert relation.subject.sourceContext is req_item1.subject
    assert relation.subject.targetContext is req_item2.subject

    disconnect(relation, relation.handles()[0])

    assert not relation.subject



"""
Implementation of interface.
"""

import gaphas

from gaphor import UML
from gaphor.core.styling import Style
from gaphor.diagram.presentation import LinePresentation, Named
from gaphor.diagram.shapes import Box, EditableText, Text
from gaphor.diagram.support import represents
from gaphor.UML.classes.interface import Folded, InterfacePort
from gaphor.UML.modelfactory import stereotypes_str


@represents(UML.Implementation)
class ImplementationItem(LinePresentation, Named):
    def __init__(self, id=None, model=None):
        super().__init__(id, model, style={"dash-style": (7.0, 5.0)})

        self.shape_middle = Box(
            Text(text=lambda: stereotypes_str(self.subject),),
            EditableText(text=lambda: self.subject.name or ""),
        )
        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")
        self._inline_style: Style = {}

    def connected_to_folded_interface(self):
        assert isinstance(self.canvas, gaphas.Canvas)
        connection = self.canvas.get_connection(self.head)
        return (
            connection
            and isinstance(connection.port, InterfacePort)
            and connection.connected.folded != Folded.NONE
        )

    def post_update(self, context):
        super().post_update(context)
        if self.connected_to_folded_interface():
            self.style["dash-style"] = ()
        else:
            self.style["dash-style"] = (7.0, 5.0)

    def draw_head(self, context):
        cr = context.cairo
        cr.move_to(0, 0)
        if context.style.get("dash-style"):
            cr.set_dash((), 0)
            cr.line_to(15, -10)
            cr.line_to(15, 10)
            cr.close_path()
            cr.stroke()
            cr.move_to(15, 0)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gaphor',
 'gaphor.SysML',
 'gaphor.SysML.blocks',
 'gaphor.SysML.blocks.tests',
 'gaphor.SysML.requirements',
 'gaphor.SysML.requirements.tests',
 'gaphor.SysML.tests',
 'gaphor.UML',
 'gaphor.UML.actions',
 'gaphor.UML.actions.tests',
 'gaphor.UML.classes',
 'gaphor.UML.classes.tests',
 'gaphor.UML.components',
 'gaphor.UML.components.tests',
 'gaphor.UML.interactions',
 'gaphor.UML.interactions.tests',
 'gaphor.UML.profiles',
 'gaphor.UML.profiles.tests',
 'gaphor.UML.states',
 'gaphor.UML.states.tests',
 'gaphor.UML.tests',
 'gaphor.UML.usecases',
 'gaphor.UML.usecases.tests',
 'gaphor.codegen',
 'gaphor.codegen.tests',
 'gaphor.core',
 'gaphor.core.modeling',
 'gaphor.core.modeling.tests',
 'gaphor.core.styling',
 'gaphor.core.styling.tests',
 'gaphor.diagram',
 'gaphor.diagram.general',
 'gaphor.diagram.general.tests',
 'gaphor.diagram.tests',
 'gaphor.plugins',
 'gaphor.plugins.console',
 'gaphor.plugins.console.tests',
 'gaphor.plugins.diagramexport',
 'gaphor.plugins.diagramexport.tests',
 'gaphor.plugins.xmiexport',
 'gaphor.plugins.xmiexport.tests',
 'gaphor.services',
 'gaphor.services.helpservice',
 'gaphor.services.tests',
 'gaphor.storage',
 'gaphor.storage.tests',
 'gaphor.tests',
 'gaphor.ui',
 'gaphor.ui.tests']

package_data = \
{'': ['*'],
 'gaphor': ['locale/*'],
 'gaphor.ui': ['icons/*', 'icons/hicolor/scalable/actions/*']}

install_requires = \
['PyGObject>=3.30,<4.0',
 'gaphas>=2.1.1,<3.0.0',
 'generic>=1.0.0,<2.0.0',
 'importlib_metadata>=1.4,<2.0',
 'pycairo>=1.18,<2.0',
 'tinycss2>=1.0.2,<2.0.0',
 'typing_extensions>=3.7.4,<4.0.0']

entry_points = \
{'console_scripts': ['gaphor = gaphor.ui:main',
                     'gaphorconvert = '
                     'gaphor.plugins.diagramexport.gaphorconvert:main'],
 'gaphor.appservices': ['app_file_manager = '
                        'gaphor.ui.appfilemanager:AppFileManager',
                        'help = gaphor.services.helpservice:HelpService',
                        'session = gaphor.services.session:Session'],
 'gaphor.modelinglanguages': ['SysML = '
                              'gaphor.SysML.modelinglanguage:SysMLModelingLanguage',
                              'UML = '
                              'gaphor.UML.modelinglanguage:UMLModelingLanguage'],
 'gaphor.services': ['component_registry = '
                     'gaphor.services.componentregistry:ComponentRegistry',
                     'consolewindow = gaphor.plugins.console:ConsoleWindow',
                     'copy = gaphor.services.copyservice:CopyService',
                     'diagram_export = '
                     'gaphor.plugins.diagramexport:DiagramExport',
                     'diagrams = gaphor.ui.mainwindow:Diagrams',
                     'element_dispatcher = '
                     'gaphor.core.modeling.elementdispatcher:ElementDispatcher',
                     'element_factory = gaphor.core.modeling:ElementFactory',
                     'elementeditor = gaphor.ui.elementeditor:ElementEditor',
                     'event_manager = gaphor.core.eventmanager:EventManager',
                     'export_menu = gaphor.ui.menufragment:MenuFragment',
                     'file_manager = gaphor.ui.filemanager:FileManager',
                     'main_window = gaphor.ui.mainwindow:MainWindow',
                     'modeling_language = '
                     'gaphor.services.modelinglanguage:ModelingLanguageService',
                     'namespace = gaphor.ui.namespace:Namespace',
                     'properties = gaphor.services.properties:Properties',
                     'recent_files = gaphor.ui.recentfiles:RecentFiles',
                     'sanitizer = gaphor.UML.sanitizerservice:SanitizerService',
                     'toolbox = gaphor.ui.toolbox:Toolbox',
                     'tools_menu = gaphor.ui.menufragment:MenuFragment',
                     'undo_manager = gaphor.services.undomanager:UndoManager',
                     'xmi_export = gaphor.plugins.xmiexport:XMIExport']}

setup_kwargs = {
    'name': 'gaphor',
    'version': '2.0.0',
    'description': 'Gaphor is the simple modeling tool written in Python.',
    'long_description': '# Gaphor <img src="logos/org.gaphor.Gaphor.svg" width="48">\n\n[![Build state](https://github.com/gaphor/gaphor/workflows/Build/badge.svg)](https://github.com/gaphor/gaphor/actions)\n[![Maintainability](https://api.codeclimate.com/v1/badges/f00974f5d7fe69fe4ecd/maintainability)](https://codeclimate.com/github/gaphor/gaphor/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/f00974f5d7fe69fe4ecd/test_coverage)](https://codeclimate.com/github/gaphor/gaphor/test_coverage)\n[![Docs build state](https://readthedocs.org/projects/gaphor/badge/?version=latest)](https://gaphor.readthedocs.io)\n[![PyPI](https://img.shields.io/pypi/v/gaphor.svg)](https://pypi.org/project/gaphor)\n[![Downloads](https://pepy.tech/badge/gaphor)](https://pepy.tech/project/gaphor)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![standard-readme compliant](https://img.shields.io/badge/readme%20style-standard-brightgreen.svg?style=flat)](https://github.com/RichardLitt/standard-readme)\n[![Gitter](https://img.shields.io/gitter/room/nwjs/nw.js.svg)](https://gitter.im/Gaphor/Lobby?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n[![All Contributors](https://img.shields.io/badge/all_contributors-24-orange.svg?style=flat-square)](#contributors)\n\n\n> Gaphor is the simple modeling tool for UML and SysML.\n\n![Gaphor Demo](docs/images/gaphor-demo.gif)\n\nGaphor is a simple and easy to use modeling tool for UML and SysML. It is aimed\nat beginning modelers who want a simple and fast tool so that they can focus on\nlearning modeling of software and systems. It is not a full featured enterprise\ntool.\n\n## :bookmark_tabs: Table of Contents\n\n- [Background](#background)\n- [Install](#floppy_disk-install)\n- [Usage](#usage)\n- [Contributing](#contributing)\n- [License](#license)\n\n## :scroll: Background\n\nGaphor is a UML and SysML modeling application written in Python. We designed\nit to be easy to use, while still being powerful. Gaphor implements a\nfully-compliant UML 2 data model, so it is much more than a picture drawing\ntool. You can use Gaphor to quickly visualize different aspects of a system as\nwell as create complete, highly complex models.\n\nGaphor is designed around the following principles:\n\n- Simplicity: The application should be easy to use. Only some basic knowledge of UML or SysML is required.\n- Consistency: UML is a graphical modeling language, so all modeling is done in a diagram.\n- Workability: The application should not bother the user every time they do something non-UML-ish.\n\nGaphor is built on [Gaphas](https://github.com/gaphor/gaphas), which provides\nthe foundational diagramming library. It is a GUI application that is built on\nGTK and cairo, [PyGObject](https://pygobject.readthedocs.io/) provides access\nto the GUI toolkit, and [PyCairo](https://pycairo.readthedocs.io/) to the 2D\ngraphics library.\n\n## :floppy_disk: Install\n\n### Windows\nTo install Gaphor on Windows you an use the [latest Gaphor.exe installer](https://github.com/gaphor/gaphor/releases).\nThere are two versions:\n1. Full Windows installation\n2. Portable installation\n\n### Linux\nTo install Gaphor in Linux use Flatpak:\n1. [Install Flatpak](https://flatpak.org/setup)\n1. `flatpak remote-add --user --if-not-exists flathub https://dl.flathub.org/repo/flathub.flatpakrepo`\n1. `flatpak install --user flathub org.gaphor.Gaphor`\n\n#### Arch Linux\n\nCan be installed from [AUR package](https://aur.archlinux.org/packages/python-gaphor/).\n\n### macOS\nWe are still working on packaging GTK with Gaphor and it is currently an\ninstallation pre-requisite.\n1. Install [homebrew](https://brew.sh)\n1. Open a terminal and execute:\n```bash\n$ brew install gobject-introspection gtk+3\n```\n\nThen install Gaphor on macOS using the [latest gaphor-macOS.dmg\ninstaller](https://github.com/gaphor/gaphor/releases).\n\n\nNote: Sometimes launching the app the first time after installation fails due\nto macOS security settings, please attempt to launch it a 2nd time if this\nhappens.\n\n### PyPI\nIf you have the latest stable version of Python installed and the Gaphor\ndependencies installed, you can also install Gaphor using a wheel from PyPI.\n\nIf you don\'t have the latest stable version of Python and the Gaphor dependencies\ninstalled, follow the development environment setup instructions in the next\nsection and instead of cloning the repository, create a\n[virtual environment](https://packaging.python.org/tutorials/installing-packages/#creating-virtual-environments),\nand then execute the following:\n\n```bash\n$ pip install gaphor\n$ gaphor\n```\n\n### Development\n\n#### Windows\n\nTo setup a development environment in Windows:\n1) Go to http://www.msys2.org/ and download the x86_64 installer\n1) Follow the instructions on the page for setting up the basic environment\n1) Run ``C:\\msys64\\mingw64.exe`` - a terminal window should pop up\n```bash\n$ pacman -Suy\n$ pacman -S git mingw-w64-x86_64-gcc mingw-w64-x86_64-gtk3 \\\nmingw-w64-x86_64-pkg-config mingw-w64-x86_64-cairo \\\nmingw-w64-x86_64-gobject-introspection mingw-w64-x86_64-python \\\nmingw-w64-x86_64-python-gobject mingw-w64-x86_64-python-cairo \\\nmingw-w64-x86_64-python-pip\n```\n\nEnsure `/mingw64/bin` is added to your `PATH`:\n```bash\n$ export PATH=/mingw64/bin:$PATH\n```\n\n[Clone the\nrepository](https://help.github.com/en/github/creating-cloning-and-archiving-repositories/cloning-a-repository).\n\n```bash\n$ cd gaphor\n$ source ./venv -S\n$ poetry run gaphor\n```\n\n#### Linux\nTo setup a development environment with Linux, you first need the latest\nstable version of Python. In order to get the latest stable version, we\nrecommend that you install [pyenv](https://github.com/pyenv/pyenv).\nInstall the pyenv\n[prerequisites](https://github.com/pyenv/pyenv/wiki/Common-build-problems#prerequisites)\nfirst, and then install pyenv:\n\n```bash\n$ curl https://pyenv.run | bash\n```\n\nMake sure you follow the instruction at the end of the installation script\nto install the commands in your shell\'s rc file. Finally install the latest\nversion of Python by executing:\n\n```bash\n$ pyenv install 3.x.x\n```\nWhere 3.x.x is replaced by the latest stable version of Python.\n\nNext install the Gaphor prerequisites by installing the gobject introspection\nand cairo build dependencies, for example, in Ubuntu execute:\n\n```bash\n$ sudo apt-get install -y python3-dev python3-gi python3-gi-cairo\n    gir1.2-gtk-3.0 libgirepository1.0-dev libcairo2-dev\n```\n[Clone the\nrepository](https://help.github.com/en/github/creating-cloning-and-archiving-repositories/cloning-a-repository).\n\n```\n$ cd gaphor\n$ source ./venv\n$ poetry run gaphor\n```\n\n#### macOS\nTo setup a development environment with macOS:\n1. Install [homebrew](https://brew.sh)\n1. Open a terminal and execute:\n```bash\n$ brew install python3 gobject-introspection gtk+3\n```\n[Clone the\nrepository](https://help.github.com/en/github/creating-cloning-and-archiving-repositories/cloning-a-repository).\n```\n$ cd gaphor\n$ source ./venv\n$ poetry run gaphor\n```\n\n## :flashlight: Usage\n### Creating models\n\nOnce Gaphor is started a new empty model is automatically created. The main\ndiagram is already open in the Diagram section.\n\nSelect an element you want to place, for example a Class, by clicking on the icon in\nthe Toolbox and click on the diagram. This will place a new\nClass item instance on the diagram and add a new Class to the model (it shows\nup in the Navigation). The selected tool will reset itself to\nthe Pointer tool if the option \'\'Diagram -> Reset tool\'\' is selected.\n\nSome elements are not directly visible. The section in the toolbox is collapsed\nand needs to be clicked first to reveal its contents.\n\nGaphor only has one diagram type, and it does not enforce which elements should\nbe placed on a diagram.\n\n### Create a New Diagram\n\n1. Use the Navigation to select an element that can contain a diagram (a\nPackage or Profile)\n1. Select Diagram, and New diagram. A new diagram is created.\n\n### Copy and Paste\n\nItems in a diagram can be copied and pasted in the same diagram or other\ndiagrams. Pasting places an existing item in the diagram, but the item itself\nis not duplicated. In other words, if you paste a Class object in a diagram,\nthe Class will be added to the diagram, but there will be no new Class in the\nNavigation.\n\n### Drag and Drop\n\nAdding an existing element to a diagram is done by dragging the element from\nthe Navigation section onto a diagram. Diagrams and attribute/operations of a\nClass show up in the Navigation but can not be added to a diagram.\n\nElements can also be dragged within the Navigation in order to rearrange them\nin to different packages.\n\n\n## :heart: Contributing\n\nThanks goes to these wonderful people ([emoji key](https://github.com/kentcdodds/all-contributors#emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="https://github.com/amolenaar"><img src="https://avatars0.githubusercontent.com/u/96249?v=4" width="100px;" alt=""/><br /><sub><b>Arjan Molenaar</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=amolenaar" title="Code">💻</a> <a href="https://github.com/gaphor/gaphor/issues?q=author%3Aamolenaar" title="Bug reports">🐛</a> <a href="https://github.com/gaphor/gaphor/commits?author=amolenaar" title="Documentation">📖</a> <a href="https://github.com/gaphor/gaphor/pulls?q=is%3Apr+reviewed-by%3Aamolenaar" title="Reviewed Pull Requests">👀</a> <a href="#question-amolenaar" title="Answering Questions">💬</a> <a href="https://github.com/gaphor/gaphor/issues?q=author%3Aamolenaar" title="Bug reports">🐛</a> <a href="#plugin-amolenaar" title="Plugin/utility libraries">🔌</a> <a href="https://github.com/gaphor/gaphor/commits?author=amolenaar" title="Tests">⚠️</a></td>\n    <td align="center"><a href="https://github.com/wrobell"><img src="https://avatars2.githubusercontent.com/u/105664?v=4" width="100px;" alt=""/><br /><sub><b>wrobell</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=wrobell" title="Code">💻</a> <a href="https://github.com/gaphor/gaphor/commits?author=wrobell" title="Tests">⚠️</a> <a href="https://github.com/gaphor/gaphor/issues?q=author%3Awrobell" title="Bug reports">🐛</a> <a href="#design-wrobell" title="Design">🎨</a></td>\n    <td align="center"><a href="https://ghuser.io/danyeaw"><img src="https://avatars1.githubusercontent.com/u/10014976?v=4" width="100px;" alt=""/><br /><sub><b>Dan Yeaw</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=danyeaw" title="Code">💻</a> <a href="https://github.com/gaphor/gaphor/commits?author=danyeaw" title="Tests">⚠️</a> <a href="https://github.com/gaphor/gaphor/commits?author=danyeaw" title="Documentation">📖</a> <a href="#platform-danyeaw" title="Packaging/porting to new platform">📦</a> <a href="#infra-danyeaw" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a> <a href="https://github.com/gaphor/gaphor/issues?q=author%3Adanyeaw" title="Bug reports">🐛</a> <a href="#question-danyeaw" title="Answering Questions">💬</a></td>\n    <td align="center"><a href="https://github.com/melisdogan"><img src="https://avatars2.githubusercontent.com/u/33630433?v=4" width="100px;" alt=""/><br /><sub><b>melisdogan</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=melisdogan" title="Documentation">📖</a></td>\n    <td align="center"><a href="http://www.boduch.ca"><img src="https://avatars2.githubusercontent.com/u/114619?v=4" width="100px;" alt=""/><br /><sub><b>Adam Boduch</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=adamboduch" title="Code">💻</a> <a href="https://github.com/gaphor/gaphor/commits?author=adamboduch" title="Tests">⚠️</a> <a href="https://github.com/gaphor/gaphor/issues?q=author%3Aadamboduch" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/egroeper"><img src="https://avatars3.githubusercontent.com/u/535113?v=4" width="100px;" alt=""/><br /><sub><b>Enno Gröper</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=egroeper" title="Code">💻</a></td>\n    <td align="center"><a href="https://pfeifle.tech"><img src="https://avatars2.githubusercontent.com/u/23027708?v=4" width="100px;" alt=""/><br /><sub><b>JensPfeifle</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=JensPfeifle" title="Documentation">📖</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="http://www.aejh.co.uk"><img src="https://avatars1.githubusercontent.com/u/927233?v=4" width="100px;" alt=""/><br /><sub><b>Alexis Howells</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=aejh" title="Documentation">📖</a></td>\n    <td align="center"><a href="http://encolpe.wordpress.com"><img src="https://avatars1.githubusercontent.com/u/124361?v=4" width="100px;" alt=""/><br /><sub><b>Encolpe DEGOUTE</b></sub></a><br /><a href="#translation-encolpe" title="Translation">🌍</a></td>\n    <td align="center"><a href="https://github.com/choff"><img src="https://avatars1.githubusercontent.com/u/309979?v=4" width="100px;" alt=""/><br /><sub><b>Christian Hoff</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=choff" title="Code">💻</a></td>\n    <td align="center"><a href="https://oskuro.net/"><img src="https://avatars3.githubusercontent.com/u/929712?v=4" width="100px;" alt=""/><br /><sub><b>Jordi Mallach</b></sub></a><br /><a href="#translation-jmallach" title="Translation">🌍</a></td>\n    <td align="center"><a href="https://github.com/tonytheleg"><img src="https://avatars3.githubusercontent.com/u/43508092?v=4" width="100px;" alt=""/><br /><sub><b>Tony</b></sub></a><br /><a href="#maintenance-tonytheleg" title="Maintenance">🚧</a></td>\n    <td align="center"><a href="https://github.com/jischebeck"><img src="https://avatars0.githubusercontent.com/u/3011242?v=4" width="100px;" alt=""/><br /><sub><b>Jan</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Ajischebeck" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="http://btibert3.github.io"><img src="https://avatars2.githubusercontent.com/u/203343?v=4" width="100px;" alt=""/><br /><sub><b>Brock Tibert</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3ABtibert3" title="Bug reports">🐛</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="http://www.rmunoz.net"><img src="https://avatars2.githubusercontent.com/u/23944?v=4" width="100px;" alt=""/><br /><sub><b>Rafael Muñoz Cárdenas</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3AMenda" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/mbessonov"><img src="https://avatars2.githubusercontent.com/u/172974?v=4" width="100px;" alt=""/><br /><sub><b>Mikhail Bessonov</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Ambessonov" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="http://twitter.com/kapilvt"><img src="https://avatars3.githubusercontent.com/u/21650?v=4" width="100px;" alt=""/><br /><sub><b>Kapil Thangavelu</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Akapilt" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/DimShadoWWW"><img src="https://avatars2.githubusercontent.com/u/25516?v=4" width="100px;" alt=""/><br /><sub><b>DimShadoWWW</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3ADimShadoWWW" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="http://nedko.arnaudov.name"><img src="https://avatars2.githubusercontent.com/u/96399?v=4" width="100px;" alt=""/><br /><sub><b>Nedko Arnaudov</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Anedko" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/Alexander-Wilms"><img src="https://avatars2.githubusercontent.com/u/3226457?v=4" width="100px;" alt=""/><br /><sub><b>Alexander Wilms</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3AAlexander-Wilms" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="http://stevenliu216.github.io"><img src="https://avatars3.githubusercontent.com/u/1274417?v=4" width="100px;" alt=""/><br /><sub><b>Steven Liu</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Astevenliu216" title="Bug reports">🐛</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="https://github.com/ruimaciel"><img src="https://avatars3.githubusercontent.com/u/169121?v=4" width="100px;" alt=""/><br /><sub><b>Rui Maciel</b></sub></a><br /><a href="#ideas-ruimaciel" title="Ideas, Planning, & Feedback">🤔</a></td>\n    <td align="center"><a href="https://github.com/ezickler"><img src="https://avatars3.githubusercontent.com/u/3604310?v=4" width="100px;" alt=""/><br /><sub><b>Enno Zickler</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Aezickler" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/tronta"><img src="https://avatars1.githubusercontent.com/u/5135577?v=4" width="100px;" alt=""/><br /><sub><b>tronta</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Atronta" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/actionless"><img src="https://avatars1.githubusercontent.com/u/1655669?v=4" width="100px;" alt=""/><br /><sub><b>Yauhen Kirylau</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=actionless" title="Documentation">📖</a> <a href="#platform-actionless" title="Packaging/porting to new platform">📦</a> <a href="#ideas-actionless" title="Ideas, Planning, & Feedback">🤔</a> <a href="https://github.com/gaphor/gaphor/issues?q=author%3Aactionless" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="http://www.zorinos.com"><img src="https://avatars1.githubusercontent.com/u/34811668?v=4" width="100px;" alt=""/><br /><sub><b>albanobattistella</b></sub></a><br /><a href="#translation-albanobattistella" title="Translation">🌍</a></td>\n    <td align="center"><a href="https://gavr123456789.gitlab.io/hugo-test/"><img src="https://avatars3.githubusercontent.com/u/30507409?v=4" width="100px;" alt=""/><br /><sub><b>gavr123456789</b></sub></a><br /><a href="#ideas-gavr123456789" title="Ideas, Planning, & Feedback">🤔</a></td>\n    <td align="center"><a href="https://github.com/Xander982"><img src="https://avatars2.githubusercontent.com/u/51178927?v=4" width="100px;" alt=""/><br /><sub><b>Xander982</b></sub></a><br /><a href="#content-Xander982" title="Content">🖋</a> <a href="https://github.com/gaphor/gaphor/issues?q=author%3AXander982" title="Bug reports">🐛</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="https://github.com/seryafarma"><img src="https://avatars0.githubusercontent.com/u/3274071?v=4" width="100px;" alt=""/><br /><sub><b>seryafarma</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/commits?author=seryafarma" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://github.com/dlagg"><img src="https://avatars3.githubusercontent.com/u/44321931?v=4" width="100px;" alt=""/><br /><sub><b>Jorge DLG</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Adlagg" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/abjurstrom-torch"><img src="https://avatars1.githubusercontent.com/u/62608984?v=4" width="100px;" alt=""/><br /><sub><b>Adam Bjurstrom</b></sub></a><br /><a href="https://github.com/gaphor/gaphor/issues?q=author%3Aabjurstrom-torch" title="Bug reports">🐛</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-enable -->\n<!-- prettier-ignore-end -->\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the\n[all-contributors](https://github.com/kentcdodds/all-contributors)\nspecification. Contributions of any kind are welcome!\n\n1.  Check for open issues or open a fresh issue to start a discussion\n    around a feature idea or a bug. There is a\n    [first-timers-only](https://github.com/gaphor/gaphor/issues?utf8=%E2%9C%93&q=is%3Aissue+is%3Aopen+label%3Afirst-timers-only)\n    tag for issues that should be ideal for people who are not very\n    familiar with the codebase yet.\n2.  Fork [the repository](https://github.com/gaphor/gaphor) on\n    GitHub to start making your changes to the **master** branch (or\n    branch off of it).\n3.  Write a test which shows that the bug was fixed or that the feature\n    works as expected.\n4.  Send a pull request and bug the maintainers until it gets merged and\n    published. :smile:\n\nSee [the contributing file](CONTRIBUTING.md)!\n\n\n## :copyright: License\nCopyright (C) Arjan Molenaar and Dan Yeaw\n\nLicensed under the [Apache License v2](LICENSE.txt).\n\nSummary: You can do what you like with Gaphor, as long as you include the\nrequired notices. This permissive license contains a patent license from the\ncontributors of the code.\n',
    'author': 'Arjan J. Molenaar',
    'author_email': 'gaphor@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gaphor.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Formatting of UML model elements into text tests.
"""

import pytest

from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import ElementFactory
from gaphor.UML import model
from gaphor.UML import uml as UML
from gaphor.UML.umlfmt import format
from gaphor.UML.umllex import parse


@pytest.fixture
def factory():
    event_manager = EventManager()
    return ElementFactory(event_manager)


def add_tag_is_foo_metadata_field(e, factory):
    s = factory.create(UML.Stereotype)
    s.ownedAttribute = factory.create(UML.Property)
    parse(s.ownedAttribute[0], "tag: str")

    instance_spec = model.apply_stereotype(e, s)
    slot = model.add_slot(instance_spec, s.ownedAttribute[0])
    slot.value = "foo"
    return slot


@pytest.mark.parametrize(
    "text,formatted_text",
    [
        ("", ""),
        ("not an attribute===foobar", "+ not an attribute===foobar"),
        ("myattr", "+ myattr"),
        ("myattr:int", "+ myattr: int"),
        ("- myattr:int[3]", "- myattr: int[3]"),
        ("- myattr:int[0..1]", "- myattr: int[0..1]"),
        ("/myattr:int", "+ /myattr: int"),
        ("myattr:int=3", "+ myattr: int = 3"),
    ],
)
def test_attribute(factory, text, formatted_text):
    """Test simple attribute formatting
    """
    a = factory.create(UML.Property)
    parse(a, text)

    assert formatted_text == format(a)


def test_attribute_with_applied_stereotype(factory):
    a = factory.create(UML.Property)
    parse(a, "myattr: int")
    add_tag_is_foo_metadata_field(a, factory)

    assert '+ myattr: int { tag = "foo" }' == format(a, tags=True)


@pytest.mark.parametrize(
    "text,name_part, mult_part",
    [
        ("", "", ""),
        ("not an association end[3]", "+ not an association end[3]", ""),
        ("myattr", "+ myattr", ""),
        ("myattr[0..1]", "+ myattr", "0..1"),
        ("- myattr[0..1]", "- myattr", "0..1"),
    ],
)
def test_association_end(factory, text, name_part, mult_part):
    """Test simple attribute formatting
    """
    a = factory.create(UML.Property)
    a.association = factory.create(UML.Association)
    parse(a, text)

    assert (name_part, mult_part) == format(a)


def test_attribute_with_type(factory):
    """Test simple attribute formatting
    """
    a = factory.create(UML.Property)
    a.type = factory.create(UML.Class)
    a.name = "attr"
    a.type.name = "MyClass"

    assert "+ attr: MyClass" == format(a)


def test_association_end_with_applied_stereotype(factory):
    a = factory.create(UML.Property)
    a.association = factory.create(UML.Association)
    parse(a, "myattr[1]")
    add_tag_is_foo_metadata_field(a, factory)

    assert ("+ myattr", '1 { tag = "foo" }') == format(a)


@pytest.mark.parametrize(
    "text,formatted_text",
    [
        ("", ""),
        ("not an operation", "+ not an operation()"),
        ("+ myoper(param: str): int", "+ myoper(in param: str): int"),
        ("+ myoper(param: str = 'aap')", "+ myoper(in param: str = 'aap')"),
        ("- myoper(out param: str): int[2]", "- myoper(out param: str): int[2]"),
        ("- myoper(out param: str): int[0..3]", "- myoper(out param: str): int[0..3]"),
        ("- myoper(p1: str[2], p2:int[*])", "- myoper(in p1: str[2], in p2: int[*])"),
        (
            "- myoper(p1: str[2], p2:int[1..*])",
            "- myoper(in p1: str[2], in p2: int[1..*])",
        ),
        ("+ (param: str): int", "+ (param: str): int"),
    ],
)
def test_operation(factory, text, formatted_text):
    """Test simple operation formatting
    """
    o = factory.create(UML.Operation)
    parse(o, text)

    assert formatted_text == format(o)


def test_slot(factory):
    a = factory.create(UML.Property)
    parse(a, "myattr: int")
    slot = add_tag_is_foo_metadata_field(a, factory)

    assert 'tag = "foo"' == format(slot)



"""
Parsing of UML model elements from string tests.
"""

import pytest

from gaphor import UML
from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import ElementFactory
from gaphor.UML.umllex import attribute_pat, operation_pat, parameter_pat


def dump_prop(prop):
    attribute_pat.match(prop)


def dump_oper(oper):
    m = operation_pat.match(oper)
    if m:
        g = m.group
    else:
        # set name to oper
        return
    if g("params"):
        params = g("params")
        while params:
            m = parameter_pat.match(params)
            g = m.group
            params = g("rest")


dump_prop("#/name")
dump_prop('+ / name : str[1..*] = "aap" { static }')
dump_prop('+ / name : str[*] = "aap" { static }')

dump_oper('myfunc(aap:str = "aap", out two): type')
dump_oper("   myfunc2 ( ): type")
dump_oper('myfunc(aap:str[1] = "aap" { tag1, tag2 }, out two {tag3}): type')


@pytest.fixture
def factory():
    event_manager = EventManager()
    return ElementFactory(event_manager)


def test_parse_property_simple(factory):
    """Test simple property parsing
    """
    a = factory.create(UML.Property)
    UML.parse(a, "myattr")
    assert not a.isDerived
    assert "myattr" == a.name
    assert a.typeValue is None, a.typeValue
    assert a.lowerValue is None, a.lowerValue
    assert a.upperValue is None, a.upperValue
    assert a.defaultValue is None, a.defaultValue


def test_parse_property_complex(factory):
    """Test complex property parsing
    """
    a = factory.create(UML.Property)

    UML.parse(a, '+ / name : str[0..*] = "aap" { static }')
    assert "public" == a.visibility
    assert a.isDerived
    assert "name" == a.name
    assert "str" == a.typeValue
    assert "0" == a.lowerValue
    assert "*" == a.upperValue
    assert '"aap"' == a.defaultValue


def test_parse_property_invalid(factory):
    """Test parsing property with invalid syntax
    """
    a = factory.create(UML.Property)

    UML.parse(a, '+ name = str[*] = "aap" { static }')
    assert '+ name = str[*] = "aap" { static }' == a.name
    assert not a.isDerived
    assert not a.typeValue
    assert not a.lowerValue
    assert not a.upperValue
    assert not a.defaultValue


def test_parse_association_end(factory):
    """Test parsing of association end
    """
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a

    UML.parse(p, "end")
    assert "end" == p.name
    assert not p.typeValue
    assert not p.lowerValue
    assert not p.upperValue
    assert not p.defaultValue


def test_parse_association_end_multiplicity(factory):
    """Test parsing of multiplicity
    """
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    UML.parse(p, "0..2 { tag }")
    assert p.name is None
    assert not p.typeValue
    assert "0" == p.lowerValue
    assert "2" == p.upperValue
    assert not p.defaultValue


def test_parse_association_end_multiplicity2(factory):
    """Test parsing of multiplicity with multiline constraints
    """
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    UML.parse(p, "0..2 { tag1, \ntag2}")
    assert p.name is None
    assert not p.typeValue
    assert "0" == p.lowerValue
    assert "2" == p.upperValue
    assert not p.defaultValue


def test_parse_association_end_derived_end(factory):
    """Test parsing derived association end
    """
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a
    UML.parse(p, "-/end[*] { mytag}")
    assert "private" == p.visibility
    assert p.isDerived
    assert "end" == p.name
    assert not p.typeValue
    assert not p.lowerValue
    assert "*" == p.upperValue
    assert not p.defaultValue


def test_parse_association_end_with_type(factory):
    """Test parsing of association end, type is ignored
    """
    a = factory.create(UML.Association)
    p = factory.create(UML.Property)
    p.association = a

    UML.parse(p, "end: TypeVal")
    assert "end" == p.name
    assert not p.typeValue
    assert not p.lowerValue
    assert not p.upperValue
    assert not p.defaultValue


def test_parse_operation(factory):
    """Test parsing simple operation
    """
    o = factory.create(UML.Operation)
    UML.parse(o, "myfunc()")
    assert "myfunc" == o.name
    assert not o.returnResult[0].typeValue
    assert not o.formalParameter


def test_parse_operation_return(factory):
    """Test parsing operation with return value
    """
    o = factory.create(UML.Operation)
    UML.parse(o, "+ myfunc(): int")
    assert "myfunc" == o.name
    assert "int" == o.returnResult[0].typeValue
    assert "public" == o.visibility
    assert not o.formalParameter


def test_parse_operation_2_params(factory):
    """Test parsing of operation with two parameters
    """
    o = factory.create(UML.Operation)
    UML.parse(o, "# myfunc2 (a: str, b: int = 3 {  static}): float")
    assert "myfunc2" == o.name
    assert "float" == o.returnResult[0].typeValue
    assert "protected" == o.visibility
    assert len(o.formalParameter) == 2
    assert "a" == o.formalParameter[0].name
    assert "str" == o.formalParameter[0].typeValue
    assert o.formalParameter[0].defaultValue is None
    assert "b" == o.formalParameter[1].name
    assert "int" == o.formalParameter[1].typeValue
    assert "3" == o.formalParameter[1].defaultValue


def test_parse_operation_1_param(factory):
    """Test parsing of operation with one parameter
    """
    o = factory.create(UML.Operation)
    UML.parse(o, "- myfunc2 (a: node): double")
    assert "myfunc2" == o.name
    assert "double" == o.returnResult[0].typeValue
    assert "private" == o.visibility
    assert len(o.formalParameter) == 1
    assert "a" == o.formalParameter[0].name
    assert "node" == o.formalParameter[0].typeValue
    assert o.formalParameter[0].defaultValue is None


def test_parse_operation_invalid_syntax(factory):
    """Test operation parsing with invalid syntax
    """
    o = factory.create(UML.Operation)
    UML.parse(o, "- myfunc2: myType2")
    assert "- myfunc2: myType2" == o.name



import pytest

from gaphor.core.modeling import NamedElement, Presentation
from gaphor.diagram.general import CommentLineItem
from gaphor.diagram.presentation import Classified, Named
from gaphor.diagram.support import get_model_element
from gaphor.SysML import diagramitems
from gaphor.UML import Classifier


def _issubclass(child, parent):
    try:
        return issubclass(child, parent)
    except TypeError:
        return False


def all_items_and_elements():
    return [
        [cls, get_model_element(cls)]
        for cls in diagramitems.__dict__.values()
        if _issubclass(cls, Presentation) and get_model_element(cls)
    ]


@pytest.mark.parametrize(
    "item_class",
    [cls for cls in diagramitems.__dict__.values() if _issubclass(cls, Presentation)],
)
def test_all_diagram_items_have_a_model_element_mapping(item_class):
    if item_class is CommentLineItem:
        assert not get_model_element(item_class)
    else:
        assert get_model_element(item_class)


NAMED_EXCLUSIONS = []


@pytest.mark.parametrize(
    "item_class,element_class", all_items_and_elements(),
)
def test_all_named_elements_have_named_items(item_class, element_class):
    if item_class in NAMED_EXCLUSIONS:
        return

    if issubclass(element_class, NamedElement):
        assert issubclass(item_class, Named)

    if issubclass(item_class, Named):
        assert issubclass(element_class, NamedElement)


CLASSIFIED_EXCLUSIONS = []


@pytest.mark.parametrize(
    "item_class,element_class", all_items_and_elements(),
)
def test_all_classifier_elements_have_classified_items(item_class, element_class):
    if item_class in CLASSIFIED_EXCLUSIONS:
        return

    if issubclass(element_class, Classifier):
        assert issubclass(item_class, Classified)

    if issubclass(item_class, Classified):
        assert issubclass(element_class, Classifier)



"""Test case that checks the working of the uml_coder.py module."""

from gaphor.application import distribution
from gaphor.codegen.override import Overrides
from gaphor.codegen.uml_coder import generate


class PseudoFile:
    def __init__(self):
        self.data = ""

    def write(self, data):
        self.data += data

    def close(self):
        pass


def test_loading():
    model_file = distribution().locate_file("test-models/test-model.gaphor")
    outfile = PseudoFile()

    generate(model_file, outfile)

    assert outfile.data == GENERATED, f'"""{outfile.data}"""'


GENERATED = """# This file is generated by codegen.py. DO NOT EDIT!

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Callable, List, Optional

from gaphor.core.modeling.properties import (
    association,
    attribute,
    derived,
    derivedunion,
    enumeration,
    redefine,
    relation_many,
    relation_one,
)
class Element:
    pass


class SubClass(Element):
    name2: relation_many[C]
    concrete: relation_many[D]
    value: attribute[str]
    abstract: relation_many[C]
    name4: relation_many[D]  # type: ignore[assignment]


class C:
    attr: attribute[str]
    name1: relation_many[SubClass]
    base: relation_many[SubClass]


class D(C):
    subbase: relation_many[SubClass]
    name3: relation_many[SubClass]


# class 'ValSpec' has been stereotyped as 'SimpleAttribute'
# class 'ShouldNotShowUp' has been stereotyped as 'SimpleAttribute' too
C.attr = attribute('attr', str)
C.name1 = association('name1', SubClass, opposite='name2')
SubClass.name2 = association('name2', C, opposite='name1')
C.base = association('base', SubClass, opposite='abstract')
D.subbase = association('subbase', SubClass, opposite='concrete')
SubClass.concrete = association('concrete', D, opposite='subbase')
D.name3 = association('name3', SubClass, opposite='name4')
# 'SubClass.value' is a simple attribute
SubClass.value = attribute('value', str)
SubClass.abstract = derivedunion(SubClass, 'abstract', C, 0, '*', SubClass.concrete)
SubClass.name4 = redefine(SubClass, 'name4', D, '*', SubClass.name2)
"""


class OverridesFile:
    def __init__(self, lines):
        self.lines_iter = iter(lines)

    def readline(self):
        try:
            return next(self.lines_iter) + "\n"
        except StopIteration:
            return None


def test_overrides():
    pf = OverridesFile(["override Transition", "placeholder"])
    overrides = Overrides()
    overrides.read_overrides(pf)

    assert overrides.has_override("Transition"), overrides.overrides
    assert overrides.derives("Transition") == (), overrides.overrides
    assert overrides.get_type("Transition") == "Any", overrides.overrides


def test_overrides_with_derived_items():
    pf = OverridesFile(["override Transition(Foo, Bar)", "placeholder"])
    overrides = Overrides()
    overrides.read_overrides(pf)

    assert overrides.has_override("Transition"), overrides.overrides
    assert overrides.derives("Transition") == ("Foo", "Bar"), overrides.overrides
    assert overrides.get_type("Transition") == "Any", overrides.overrides


def test_overrides_of_features_with_derived_items():
    pf = OverridesFile(["override Transition.foo(Bar.baz)", "placeholder"])
    overrides = Overrides()
    overrides.read_overrides(pf)

    assert overrides.has_override("Transition.foo"), overrides.overrides
    assert overrides.derives("Transition.foo") == ("Bar.baz",), overrides.overrides
    assert overrides.get_type("Transition") == "Any", overrides.overrides


def test_overrides_with_type():
    pf = OverridesFile(["override Transition: Foo[str, str]", "placeholder"])
    overrides = Overrides()
    overrides.read_overrides(pf)

    assert overrides.has_override("Transition"), overrides.overrides
    assert overrides.derives("Transition") == (), overrides.overrides
    assert overrides.get_type("Transition") == "Foo[str, str]", overrides.overrides


def test_overrides_with_type_with_quotes():
    pf = OverridesFile(['override Transition: Foo["Type"]', "placeholder"])
    overrides = Overrides()
    overrides.read_overrides(pf)

    assert overrides.has_override("Transition"), overrides.overrides
    assert overrides.derives("Transition") == (), overrides.overrides
    assert overrides.get_type("Transition") == 'Foo["Type"]', overrides.overrides


def test_overrides_with_derived_items_and_type():
    pf = OverridesFile(["override Transition(Foo, Bar): Baz[str, str]", "placeholder"])
    overrides = Overrides()
    overrides.read_overrides(pf)

    assert overrides.has_override("Transition"), overrides.overrides
    assert overrides.derives("Transition") == ("Foo", "Bar"), overrides.overrides
    assert overrides.get_type("Transition") == "Baz[str, str]", overrides.overrides



"""Defines a status window class for displaying the progress of
a queue."""

from gi.repository import Gdk, GLib, Gtk, Pango

from gaphor.ui.gidlethread import QueueEmpty


class StatusWindow:
    """Create a borderless window on the parent, usually the main window,
    with a label and a progress bar.  The progress bar is updated as the
    queue is updated."""

    def __init__(self, title, message, parent=None, queue=None, display=True):
        """Create the status window.  The title parameter is the title of the
        window.  The message parameter is a string displayed near the progress
        bar to indicate what is happening.  The parent parameter is the
        parent window to display the window in.  The queue parameter is a
        queue that is used to update the progress bar.  The display parameter
        will display the window if true.  This is the default."""

        self.title = title
        self.message = message
        self.parent = parent
        self.queue = queue

        self.init_window()

        if display:
            self.display()

    def init_window(self):
        """Create the window GUI component.  This will set the window and
        progress bar attributes so they can be referenced later."""

        frame = Gtk.Frame()
        vbox = Gtk.VBox(spacing=12)
        label = Gtk.Label(label=self.message)

        self.progress_bar = Gtk.ProgressBar()
        self.window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)

        self.window.set_title(self.title)
        self.window.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.window.set_transient_for(self.parent)
        self.window.set_modal(True)
        self.window.set_resizable(False)
        self.window.set_decorated(False)
        self.window.set_type_hint(Gdk.WindowTypeHint.SPLASHSCREEN)
        self.window.add(frame)

        self.progress_bar.set_size_request(400, -1)

        frame.add(vbox)

        label.set_ellipsize(Pango.EllipsizeMode.MIDDLE)

        vbox.set_border_width(12)
        vbox.pack_start(label, True, True, 0)
        vbox.pack_start(self.progress_bar, expand=False, fill=False, padding=0)

    def display(self):
        """Display the status window.  If a queue has been supplied to the
        StatusWindow instance, a new gobject idle handle is created.  Once
        the window is destroyed, this handler is removed."""

        self.window.show_all()

        if self.queue:
            self.idle_id = GLib.idle_add(
                progress_idle_handler,
                self.progress_bar,
                self.queue,
                priority=GLib.PRIORITY_LOW,
            )
            self.window.connect("destroy", remove_idle_handler, self.idle_id)

    def destroy(self):
        """Destroy the status window.  This will also remove the gobject
        handler."""

        self.window.destroy()


def progress_idle_handler(progress_bar, queue):
    """This is a gobject idle handler that updates the supplied progress bar.
    The percentage is retrieved from the queue until it is empty.  The progress
    bar is then updated with the current percentage."""

    percentage = 0
    try:
        while True:
            percentage = queue.get()
    except QueueEmpty:
        pass
    if percentage:
        progress_bar.set_fraction(min(percentage, 100.0) / 100.0)
    return True


def remove_idle_handler(window, idle_id):
    """This removes the supplied gobject idle id.  This handle is required
    by StatusWindow when it is destroyed."""

    GLib.source_remove(idle_id)



import pytest
from gi.repository import Gio, GLib

from gaphor.abc import ActionProvider
from gaphor.action import action
from gaphor.services.componentregistry import ComponentRegistry
from gaphor.ui.actiongroup import (
    apply_application_actions,
    as_variant_type,
    from_variant,
    to_variant,
    window_action_group,
)


class DummyActionProvider(ActionProvider):
    def __init__(self):
        self.quit_called = False
        self.toggle_state = None

    @action(name="app.quit")
    def quit(self):
        self.quit_called = True

    @action(name="new")
    def new(self):
        pass

    @action(name="toggle", state=True)
    def toggle(self, state):
        self.toggle_state = state

    def dyn_state(self):
        return "my-state"

    @action(name="my-action", state=dyn_state)
    def my_state(self, state):
        pass


@pytest.fixture
def dummy_action_provider():
    return DummyActionProvider()


@pytest.fixture
def component_registry(dummy_action_provider):
    component_registry = ComponentRegistry()
    component_registry.register("dummy_action_provider", dummy_action_provider)
    return component_registry


def test_window_action_group(component_registry):
    action_group, accel_group = window_action_group(component_registry)

    assert action_group.lookup_action("new")
    assert not action_group.lookup_action("quit")


def test_application_actions(component_registry):
    action_group = apply_application_actions(
        component_registry, Gio.SimpleActionGroup.new()
    )

    assert not action_group.lookup_action("new")
    assert action_group.lookup_action("quit")


def test_activate_application_action(component_registry, dummy_action_provider):

    action_group = apply_application_actions(
        component_registry, Gio.SimpleActionGroup.new()
    )
    action_group.lookup_action("quit").activate(None)

    assert dummy_action_provider.quit_called


def test_activate_toggle_action(component_registry, dummy_action_provider):
    action_group, accel_group = window_action_group(component_registry)
    action_group.lookup_action("toggle").change_state(GLib.Variant.new_boolean(True))

    assert dummy_action_provider.toggle_state is True


def test_variant_type_conversion():
    assert as_variant_type(None) is None
    assert as_variant_type(str).equal(GLib.VariantType.new("s"))
    assert as_variant_type(int).equal(GLib.VariantType.new("i"))
    assert as_variant_type(bool).equal(GLib.VariantType.new("b"))


def test_invalid_type_to_variant_type():
    with pytest.raises(ValueError):
        as_variant_type(float)


def test_python_to_variant():
    assert to_variant(None) is None
    assert to_variant("text") == GLib.Variant.new_string("text")
    assert to_variant(123) == GLib.Variant.new_int32(123)
    assert to_variant(True) == GLib.Variant.new_boolean(True)
    assert to_variant(False) == GLib.Variant.new_boolean(False)


def test_invalid_python_to_variant():
    with pytest.raises(ValueError):
        to_variant(1.0)


def test_from_variant_to_python_value():
    assert from_variant(None) is None
    assert from_variant(GLib.Variant.new_string("text")) == "text"
    assert from_variant(GLib.Variant.new_int32(123)) == 123
    assert from_variant(GLib.Variant.new_boolean(True)) is True


def test_invalid_gvariant_to_python():
    with pytest.raises(ValueError):
        from_variant(GLib.Variant.new_double(1.0))


def test_dynamic_state(component_registry):
    action_group, accel_group = window_action_group(component_registry)

    assert action_group.lookup_action("my-action")
    assert (
        from_variant(action_group.lookup_action("my-action").get_state()) == "my-state"
    )
